/* $Id: cpl_error.h,v 1.47 2007/02/01 12:49:10 llundin Exp $
 *
 * This file is part of the ESO Common Pipeline Library
 * Copyright (C) 2001-2004 European Southern Observatory
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/*
 * $Author: llundin $
 * $Date: 2007/02/01 12:49:10 $
 * $Revision: 1.47 $
 * $Name: cpl-3_1 $
 */

#ifndef CPL_ERROR_H
#define CPL_ERROR_H

#include <stdlib.h>
#include <cpl_macros.h>
#include <cpl_func.h>
 
CPL_BEGIN_DECLS

#ifndef __LINE__
#define __LINE__ 0
#endif

#ifndef __FILE__
#define __FILE__ ""
#endif


/*
    cpl_error_set, cpl_error_set_where & cpl_assure are macros
    so __FILE__ & __LINE__ are useful
 */

/**
 * @ingroup cpl_error
 * @brief
 *   Set function name, source file and line number where a CPL error occurred.
 *
 * @param function  Character string with function name
 *
 * @return Nothing.
 *
 * Set function name, source file and line number where a CPL error occurred.
 * If no error was previously set, the error is set to CPL_ERROR_UNSPECIFIED.
 *
 * @hideinitializer
 */

#define cpl_error_set_where(function)                           \
       cpl_error_set_where_macro(function, __FILE__, __LINE__)

/**
 * @ingroup cpl_error
 * @brief
 *   Set CPL error code, function name, source file and line number where 
 *   an error occurred.
 *
 * @param function  Character string with function name
 * @param code      Error code
 *
 * @return The specified error code.
 *
 * Set CPL error code, function name, source file and line number where
 * an error occurred.
 *
 * @hideinitializer
 */

#define cpl_error_set(function, code) \
                     cpl_error_set_macro(function, code, __FILE__, __LINE__)


typedef struct _cpl_error_ cpl_error;

/*
 *
 * This module provides functions to maintain the @em cpl_error status.
 */

/**
 * @ingroup cpl_error
 * @brief 
 *   Available error codes.

     @c CPL_ERROR_NONE is equal to zero and compares to less than all other
     error codes.

     All error codes compare to less than @c CPL_ERROR_EOL,
     which is of type @c cpl_error_code, but which is not usable in the CPL
     error handling. @c CPL_ERROR_EOL (End Of List) allows a new error handling
     system to be built on top of the CPL error handling system.

 */

enum _cpl_error_code_ {
    CPL_ERROR_NONE = 0,         
    /**< No error condition */
    CPL_ERROR_UNSPECIFIED = 1,
    /**< An unspecified error. FIXME: Do not set to 1 and verify correctness */
    CPL_ERROR_DUPLICATING_STREAM,
    /**< Could not duplicate output stream */
    CPL_ERROR_ASSIGNING_STREAM,
    /**< Could not associate a stream with a file descriptor */
    CPL_ERROR_FILE_IO,          
    /**< Permission denied */
    CPL_ERROR_BAD_FILE_FORMAT,  
    /**< Input file had not the expected format */
    CPL_ERROR_FILE_ALREADY_OPEN,
    /**< Attempted to open a file twice */
    CPL_ERROR_FILE_NOT_CREATED, 
    /**< Could not create a file */
    CPL_ERROR_FILE_NOT_FOUND,   
    /**< A specified file or directory was not found */
    CPL_ERROR_DATA_NOT_FOUND,           
    /**< Data searched within a valid object were not found */
    CPL_ERROR_ACCESS_OUT_OF_RANGE,      
    /**< Data were accessed beyond boundaries */
    CPL_ERROR_NULL_INPUT,               
    /**< A @c NULL pointer was found where a valid pointer was expected */
    CPL_ERROR_INCOMPATIBLE_INPUT,       
    /**< Data that had to be processed together did not match */
    CPL_ERROR_ILLEGAL_INPUT,            
    /**< Illegal values were detected */
    CPL_ERROR_ILLEGAL_OUTPUT,           
    /**< A given operation would have generated an illegal object */
    CPL_ERROR_UNSUPPORTED_MODE,         
    /**< The requested functionality is not supported */
    CPL_ERROR_SINGULAR_MATRIX,          
    /**< Could not invert a matrix */
    CPL_ERROR_DIVISION_BY_ZERO,         
    /**< Attempted to divide a number by zero */
    CPL_ERROR_TYPE_MISMATCH,             
    /**< Data were not of the expected type */
    CPL_ERROR_INVALID_TYPE,         
    /**< Data type was unsupported or invalid */
    CPL_ERROR_CONTINUE,
    /**< An iterative process did not converge */
    CPL_ERROR_EOL
    /**< To permit extensibility of error handling.
         It is a coding error to use this within CPL itself! */
};

/**
 * @ingroup cpl_error
 * @brief
 *   The cpl_error_code type definition
 */

typedef enum _cpl_error_code_ cpl_error_code;

void            cpl_error_reset(void);
void            cpl_error_set_where_macro(const char *, const char *, unsigned);
cpl_error_code  cpl_error_set_macro(const char *, cpl_error_code, const char *,
                                    unsigned);
cpl_error_code  cpl_error_get_code(void);
const char     *cpl_error_get_message(void);
const char     *cpl_error_get_function(void);
const char     *cpl_error_get_file(void);
unsigned        cpl_error_get_line(void);
const char     *cpl_error_get_where(void);

#define cpl_ensure(BOOL, ERRCODE, RETURN)                                      \
  do if (!(BOOL)) {                                                            \
     /* Evaluate exactly once using a name-space protected variable name */    \
    const cpl_error_code cpl_ensure_error = (ERRCODE);                         \
    (void) cpl_error_set(cpl_func,                                             \
                 cpl_ensure_error ? cpl_ensure_error : CPL_ERROR_UNSPECIFIED); \
    return RETURN;                                                             \
  } while (0)

#define cpl_ensure_code(BOOL, ERRCODE)                                         \
  do if (!(BOOL)) {                                                            \
     /* Evaluate exactly once using a name-space protected variable name */    \
    const cpl_error_code cpl_ensure_error = (ERRCODE);                         \
    return cpl_error_set(cpl_func,                                             \
                 cpl_ensure_error ? cpl_ensure_error : CPL_ERROR_UNSPECIFIED); \
  } while (0)

CPL_END_DECLS

#endif
/* end of cpl_error.h */
