/* $Id: cpl_image_basic.c,v 1.124 2006/11/27 11:31:46 yjung Exp $
 *
 * This file is part of the ESO Common Pipeline Library
 * Copyright (C) 2001-2004 European Southern Observatory
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/*
 * $Author: yjung $
 * $Date: 2006/11/27 11:31:46 $
 * $Revision: 1.124 $
 * $Name: cpl-3_1 $
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

/*-----------------------------------------------------------------------------
                                   Includes
 -----------------------------------------------------------------------------*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <math.h>
#include <assert.h>

#include "cpl_memory.h"
#include "cpl_vector.h"
#include "cpl_image_basic.h"
#include "cpl_image_stats.h"
#include "cpl_stats.h"
#include "cpl_image_bpm.h"
#include "cpl_mask.h"
#include "cpl_tools.h"
#include "cpl_error_impl.h"

#include "cpl_image_defs.h"

/*-----------------------------------------------------------------------------
                                   Defines
 -----------------------------------------------------------------------------*/

#define CPL_IMAGE_BASIC_ASSIGN              1
#define CPL_IMAGE_BASIC_ASSIGN_LOCAL        2
#define CPL_IMAGE_BASIC_THRESHOLD           5
#define CPL_IMAGE_BASIC_ABS                 6
#define CPL_IMAGE_BASIC_AVERAGE             7
#define CPL_IMAGE_BASIC_SUBMIN              8
#define CPL_IMAGE_BASIC_EXTRACT             9
#define CPL_IMAGE_BASIC_EXTRACTROW          10
#define CPL_IMAGE_BASIC_EXTRACTCOL          11
#define CPL_IMAGE_BASIC_COLLAPSE            12
#define CPL_IMAGE_BASIC_COLLAPSE_MEDIAN     13
#define CPL_IMAGE_BASIC_ROTATE_INT_LOCAL    14
#define CPL_IMAGE_BASIC_SHIFT_INT_LOCAL     15
#define CPL_IMAGE_BASIC_COPY                16
#define CPL_IMAGE_BASIC_FLIP_LOCAL          17
#define CPL_IMAGE_BASIC_MOVE_PIXELS         18
#define CPL_IMAGE_CHK_POSITIVE              19
#define CPL_IMAGE_CHK_INTEGER               20
#define CPL_IMAGE_CHK_NON_ZERO              21
#define CPL_IMAGE_CHK_NON_NEGATIVE          22
#define CPL_IMAGE_BASIC_OP_SCALAR           23
#define CPL_IMAGE_BASIC_SQRT                24

#define CPL_IMAGE_ADDITION(a,b,c)           a = (b) + (c)
#define CPL_IMAGE_ADDITIONASSIGN(a,b)       a += (b)
#define CPL_IMAGE_SUBTRACTION(a,b,c)        a = (b) - (c)
#define CPL_IMAGE_SUBTRACTIONASSIGN(a,b)    a -= (b)
#define CPL_IMAGE_MULTIPLICATION(a,b,c)     a = (b) * (c)
#define CPL_IMAGE_MULTIPLICATIONASSIGN(a,b) a *= (b)
  /* Caveat: The result is evaluated twice here */
  /* Caveat: Any domain checking must have been done prior to this call */
#define CPL_IMAGE_POWASSIGN(a,b) a = pow(a, b)
  /* Caveat: The result is evaluated twice here */
  /* Caveat: Any domain checking must have been done prior to this call */
#define CPL_IMAGE_EXPASSIGN(a,b) a = pow(b, a)
  /* Caveat: The result is evaluated twice here */
  /* Caveat: Any domain checking must have been done prior to this call */
#define CPL_IMAGE_LOGASSIGN(a,b) \
  do { if (a > 0) a = log(a)/log(b); \
       else a = 0; } while (0)

  /* Caveat: Any domain checking must have been done prior to this call */
#define CPL_IMAGE_DIVISION(a,b,c) \
  do { const double tc = c; \
    if (tc != 0.0) { \
      a = (b)/tc; \
    } else { \
      a = 0.0 * (b); /*Evaluate (b) once */ } \
    } while (0)

  /* Caveat: The dividend is evaluated twice here */
  /* Caveat: Any domain checking must have been done prior to this call */
#define CPL_IMAGE_DIVISIONASSIGN(a,b) \
  do { const double tb = b; \
    if (tb != 0) { a /= tb; } else { \
      a = 0; } \
  } while (0)

/*-----------------------------------------------------------------------------
                            Function prototypes
 -----------------------------------------------------------------------------*/

static double cpl_vector_get_noise(const cpl_vector *, int) ;
static double cpl_vector_get_fwhm(const cpl_vector *, int, double) ;
static cpl_error_code cpl_fft(double *, double *, const unsigned *, int, int);

/*----------------------------------------------------------------------------*/
/**
 * @defgroup cpl_image_basic Images basic operation routines
 *
 * This module provides functions to handle basic image processing functions
 * like simple image operations, thresholding, image collapsing,
 * extraction, etc...
 *
 * @par Synopsis:
 * @code
 *   #include "cpl_image_basic.h"
 * @endcode
 */
/*----------------------------------------------------------------------------*/
/**@{*/

/*-----------------------------------------------------------------------------
                            Function codes
 -----------------------------------------------------------------------------*/

#define CPL_OPERATION CPL_IMAGE_BASIC_ASSIGN
/*----------------------------------------------------------------------------*/
/**
  @brief    Add two images.
  @param    image1  first operand
  @param    image2  second operand
  @return   1 newly allocated image or NULL in error case

  Creates a new image, being the result of the operation, and returns it to
  the caller. The returned image must be deallocated using cpl_image_delete().
  The function supports images with different types among CPL_TYPE_INT, 
  CPL_TYPE_FLOAT and CPL_TYPE_DOUBLE. The returned image type is the one of the 
  first passed image.

  The bad pixels map of the result is the union of the bad pixels maps of
  the input images.
  
  Possible #_cpl_error_code_ set in this function:
  - CPL_ERROR_NULL_INPUT if (one of) the input pointer(s) is NULL
  - CPL_ERROR_ILLEGAL_INPUT if the input images have different sizes
  - CPL_ERROR_TYPE_MISMATCH if the passed image type is not supported
 */
/*----------------------------------------------------------------------------*/
cpl_image * cpl_image_add_create(
        const cpl_image *   image1, 
        const cpl_image *   image2) 
{
#define CPL_OPERATOR CPL_IMAGE_ADDITION
#include "cpl_image_basic_body.h"
#undef CPL_OPERATOR
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Subtract two images.
  @param    image1  first operand
  @param    image2  second operand
  @return   1 newly allocated image or NULL in error case
  @see      cpl_image_add_create()
 */
/*----------------------------------------------------------------------------*/
cpl_image * cpl_image_subtract_create(
        const cpl_image *   image1, 
        const cpl_image *   image2) 
{
#define CPL_OPERATOR CPL_IMAGE_SUBTRACTION
#include "cpl_image_basic_body.h"
#undef CPL_OPERATOR
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Multiply two images.
  @param    image1  first operand
  @param    image2  second operand
  @return   1 newly allocated image or NULL in error case
  @see      cpl_image_add_create()
 */
/*----------------------------------------------------------------------------*/
cpl_image * cpl_image_multiply_create(
        const cpl_image *   image1, 
        const cpl_image *   image2)
{
#define CPL_OPERATOR CPL_IMAGE_MULTIPLICATION
#include "cpl_image_basic_body.h"
#undef CPL_OPERATOR
}
 
/*----------------------------------------------------------------------------*/
/**
  @brief    Divide two images.
  @param    image1  first operand
  @param    image2  second operand
  @return   1 newly allocated image or NULL in error case
  @see      cpl_image_add_create()
  @see      cpl_image_divide()
 */
/*----------------------------------------------------------------------------*/
cpl_image * cpl_image_divide_create(
        const cpl_image *   image1, 
        const cpl_image *   image2)
{

#define CPL_OPERATOR CPL_IMAGE_DIVISION
#include "cpl_image_basic_body.h"
#undef CPL_OPERATOR
}
#undef CPL_OPERATION

#define CPL_OPERATION CPL_IMAGE_BASIC_ASSIGN_LOCAL
/*----------------------------------------------------------------------------*/
/**
  @brief    Add two images, store the result in the first image.
  @param    im1     first operand.
  @param    im2     second operand.
  @return   the #_cpl_error_code_ or CPL_ERROR_NONE
  
  The first input image is modified to contain the results of the operation.
  The function supports images with different types among CPL_TYPE_INT, 
  CPL_TYPE_FLOAT and CPL_TYPE_DOUBLE.

  The bad pixels map of the first image becomes the union of the bad pixels 
  maps of the input images.
  
  Possible #_cpl_error_code_ set in this function:
  - CPL_ERROR_NULL_INPUT if (one of) the input pointer(s) is NULL
  - CPL_ERROR_ILLEGAL_INPUT if the input images have different sizes
  - CPL_ERROR_TYPE_MISMATCH if the passed image type is not supported
 */
/*----------------------------------------------------------------------------*/
cpl_error_code cpl_image_add(
        cpl_image       *   im1, 
        const cpl_image *   im2)
{
#define CPL_OPERATOR CPL_IMAGE_ADDITIONASSIGN
#include "cpl_image_basic_body.h"
#undef CPL_OPERATOR
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Subtract two images, store the result in the first image.
  @param    im1     first operand.
  @param    im2     second operand.
  @return   the #_cpl_error_code_ or CPL_ERROR_NONE
  @see      cpl_image_add()
 */
/*----------------------------------------------------------------------------*/
cpl_error_code cpl_image_subtract(
        cpl_image       *   im1, 
        const cpl_image *   im2)
{
#define CPL_OPERATOR CPL_IMAGE_SUBTRACTIONASSIGN
#include "cpl_image_basic_body.h"
#undef CPL_OPERATOR
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Multiply two images, store the result in the first image.
  @param    im1     first operand.
  @param    im2     second operand.
  @return   the #_cpl_error_code_ or CPL_ERROR_NONE
  @see      cpl_image_add()
 */
/*----------------------------------------------------------------------------*/
cpl_error_code cpl_image_multiply(
        cpl_image       *   im1, 
        const cpl_image *   im2)
{
#define CPL_OPERATOR CPL_IMAGE_MULTIPLICATIONASSIGN
#include "cpl_image_basic_body.h"
#undef CPL_OPERATOR
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Divide two images, store the result in the first image.
  @param    im1     first operand.
  @param    im2     second operand.
  @return   the #_cpl_error_code_ or CPL_ERROR_NONE
  @see      cpl_image_add()
  @note
     The result of division with a zero-valued pixel is zero. Detection of
     division by zero must be done by the caller.
 */
/*----------------------------------------------------------------------------*/
cpl_error_code cpl_image_divide(
        cpl_image       *   im1, 
        const cpl_image *   im2)
{

#define CPL_OPERATOR CPL_IMAGE_DIVISIONASSIGN
#include "cpl_image_basic_body.h"
#undef CPL_OPERATOR
}
#undef CPL_OPERATION

#define CPL_OPERATION CPL_IMAGE_BASIC_OP_SCALAR
#define CPL_OPERATOR CPL_IMAGE_ADDITIONASSIGN
/*----------------------------------------------------------------------------*/
/**
  @brief    Elementwise addition of a scalar to an image
  @param    image   Image to be modified in place.
  @param    addend  Number to add
  @return   CPL_ERROR_NONE or the relevant the #_cpl_error_code_ on error
 
  Modifies the image by adding a number to each of its pixels.

  The operation is always performed in double precision, with a final
  cast of the result to the target image type.
  
  Images can be CPL_TYPE_INT, CPL_TYPE_FLOAT, CPL_TYPE_DOUBLE.

  Possible #_cpl_error_code_ set in this function:
  - CPL_ERROR_NULL_INPUT if (one of) the input pointer(s) is NULL
  - CPL_ERROR_TYPE_MISMATCH if the passed image type is not supported
 */
/*----------------------------------------------------------------------------*/
cpl_error_code cpl_image_add_scalar(
        cpl_image   *   image,
        double          addend)
{
    const double scalar = addend;
    int          i ;

    /* Error handling   */
    cpl_ensure_code(image != NULL, CPL_ERROR_NULL_INPUT);
    
    assert( image->pixels );
    
    /* Switch on image type */
    switch (image->type) {
#define CPL_CLASS CPL_CLASS_DOUBLE
#include "cpl_image_basic_body.h"
#undef CPL_CLASS
#define CPL_CLASS CPL_CLASS_FLOAT
#include "cpl_image_basic_body.h"
#undef CPL_CLASS
#define CPL_CLASS CPL_CLASS_INT
#include "cpl_image_basic_body.h"
#undef CPL_CLASS
      default:
        cpl_ensure_code(0, CPL_ERROR_TYPE_MISMATCH);
    }
    return CPL_ERROR_NONE ;
}
#undef CPL_OPERATOR

#define CPL_OPERATOR CPL_IMAGE_SUBTRACTIONASSIGN
/*----------------------------------------------------------------------------*/
/**
  @brief    Elementwise subtraction of a scalar from an image
  @param    image       Image to be modified in place.
  @param    subtrahend  Number to subtract
  @return   CPL_ERROR_NONE or the relevant the #_cpl_error_code_ on error
  @see      cpl_image_add_scalar()
 */
/*----------------------------------------------------------------------------*/
cpl_error_code cpl_image_subtract_scalar(
        cpl_image   *   image,
        double          subtrahend)
{
    const double scalar = subtrahend;
    int          i ;

    /* Error handling   */
    cpl_ensure_code(image != NULL, CPL_ERROR_NULL_INPUT);
    
    assert( image->pixels );
    
    /* Switch on image type */
    switch (image->type) {
#define CPL_CLASS CPL_CLASS_DOUBLE
#include "cpl_image_basic_body.h"
#undef CPL_CLASS
#define CPL_CLASS CPL_CLASS_FLOAT
#include "cpl_image_basic_body.h"
#undef CPL_CLASS
#define CPL_CLASS CPL_CLASS_INT
#include "cpl_image_basic_body.h"
#undef CPL_CLASS
      default:
        cpl_ensure_code(0, CPL_ERROR_TYPE_MISMATCH);
    }
    return CPL_ERROR_NONE ;
}
#undef CPL_OPERATOR

#define CPL_OPERATOR CPL_IMAGE_MULTIPLICATIONASSIGN
/*----------------------------------------------------------------------------*/
/**
  @brief    Elementwise multiplication of an image with a scalar
  @param    image   Image to be modified in place.
  @param    factor  Number to multiply with
  @return   CPL_ERROR_NONE or the relevant the #_cpl_error_code_ on error
  @see      cpl_image_add_scalar()
 */
/*----------------------------------------------------------------------------*/
cpl_error_code cpl_image_multiply_scalar(
        cpl_image   *   image,
        double          factor)
{
    const double scalar = factor;
    int          i ;

    /* Error handling   */
    cpl_ensure_code(image != NULL, CPL_ERROR_NULL_INPUT);
    
    assert( image->pixels );
    
    /* Switch on image type */
    switch (image->type) {
#define CPL_CLASS CPL_CLASS_DOUBLE
#include "cpl_image_basic_body.h"
#undef CPL_CLASS
#define CPL_CLASS CPL_CLASS_FLOAT
#include "cpl_image_basic_body.h"
#undef CPL_CLASS
#define CPL_CLASS CPL_CLASS_INT
#include "cpl_image_basic_body.h"
#undef CPL_CLASS
      default:
        cpl_ensure_code(0, CPL_ERROR_TYPE_MISMATCH);
    }
    return CPL_ERROR_NONE ;
}
#undef CPL_OPERATOR

/*----------------------------------------------------------------------------*/
/**
  @brief    Elementwise division of an image with a scalar
  @param    image   Image to be modified in place.
  @param    divisor  Non-zero number to divide with
  @return   CPL_ERROR_NONE or the relevant the #_cpl_error_code_ on error
 
  Modifies the image by dividing each of its pixels with a number.

  Images can be CPL_TYPE_INT, CPL_TYPE_FLOAT, CPL_TYPE_DOUBLE.

  If divisor is zero, the image is not modified (and an error is returned).

  Possible #_cpl_error_code_ set in this function:
  - CPL_ERROR_NULL_INPUT if (one of) the input pointer(s) is NULL
  - CPL_ERROR_TYPE_MISMATCH if the passed image type is not supported
  - CPL_ERROR_DIVISION_BY_ZERO a division by 0 occurs
 */
/*----------------------------------------------------------------------------*/
cpl_error_code cpl_image_divide_scalar(
        cpl_image   *   image,
        double          divisor)
{
#ifndef CPL_DIVIDE_SLOPPY
#define CPL_OPERATOR CPL_IMAGE_DIVISIONASSIGN
    const double scalar = divisor;
#else
#define CPL_OPERATOR CPL_IMAGE_MULTIPLICATIONASSIGN
    const double scalar = divisor == 0 ? 0 : 1/divisor ;
#endif

    int          i ;

    /* Error handling   */
    cpl_ensure_code(image   != NULL, CPL_ERROR_NULL_INPUT);
    cpl_ensure_code(divisor != 0,    CPL_ERROR_DIVISION_BY_ZERO);
    
    assert( image->pixels );
    
    /* Switch on image type */
    switch (image->type) {
#define CPL_CLASS CPL_CLASS_DOUBLE
#include "cpl_image_basic_body.h"
#undef CPL_CLASS
#define CPL_CLASS CPL_CLASS_FLOAT
#include "cpl_image_basic_body.h"
#undef CPL_CLASS
#define CPL_CLASS CPL_CLASS_INT
#include "cpl_image_basic_body.h"
#undef CPL_CLASS
      default:
        cpl_ensure_code(0, CPL_ERROR_TYPE_MISMATCH);
    }
    return CPL_ERROR_NONE ;
}
#undef CPL_OPERATOR
#undef CPL_OPERATION

/*----------------------------------------------------------------------------*/
/**
  @brief    Compute the elementwise logarithm of the image.
  @param    image   Image to be modified in place.
  @param    base    Base of the logarithm.
  @return   CPL_ERROR_NONE or the relevant the #_cpl_error_code_ on error
 
  Modifies the image by computing the base-scalar logarithm of each of its
  pixels.

  Images can be CPL_TYPE_INT, CPL_TYPE_FLOAT or CPL_TYPE_DOUBLE.

  The base and all the pixels must be positive and the base must be different
  from 1, or a cpl_error_code will be returned and the image will be left
  unmodified.

  Possible #_cpl_error_code_ set in this function:
  - CPL_ERROR_NULL_INPUT if (one of) the input pointer(s) is NULL
  - CPL_ERROR_ILLEGAL_INPUT if base is negative or if the image contains
    any negative value.
  - CPL_ERROR_TYPE_MISMATCH if the passed image type is not supported
  - CPL_ERROR_DIVISION_BY_ZERO a division by 0 occurs
 */
/*----------------------------------------------------------------------------*/
cpl_error_code cpl_image_logarithm(
        cpl_image   *   image,
        double          base)
{
   /* Assume that the compiler exploits that log(base) is constant */
    const double scalar = base;
    int          i ;
    unsigned char ok = 1;

    /* Error handling   */
    cpl_ensure_code(image  != NULL, CPL_ERROR_NULL_INPUT);
    cpl_ensure_code(base > 0,       CPL_ERROR_ILLEGAL_INPUT);
    cpl_ensure_code(base != 1,      CPL_ERROR_DIVISION_BY_ZERO);

    assert( image->pixels );
    
#define CPL_OPERATION CPL_IMAGE_CHK_POSITIVE
    /* Switch on image type */
    switch (image->type) {
#define CPL_CLASS CPL_CLASS_DOUBLE
#include "cpl_image_basic_body.h"
#undef CPL_CLASS
#define CPL_CLASS CPL_CLASS_FLOAT
#include "cpl_image_basic_body.h"
#undef CPL_CLASS
#define CPL_CLASS CPL_CLASS_INT
#include "cpl_image_basic_body.h"
#undef CPL_CLASS
#undef CPL_OPERATION
    default:
        cpl_ensure_code(0, CPL_ERROR_TYPE_MISMATCH);
    }
    cpl_ensure_code(ok, CPL_ERROR_ILLEGAL_INPUT);

#define CPL_OPERATION CPL_IMAGE_BASIC_OP_SCALAR
#define CPL_OPERATOR CPL_IMAGE_LOGASSIGN
    /* Switch on image type */
    switch (image->type) {
#define CPL_CLASS CPL_CLASS_DOUBLE
#include "cpl_image_basic_body.h"
#undef CPL_CLASS
#define CPL_CLASS CPL_CLASS_FLOAT
#include "cpl_image_basic_body.h"
#undef CPL_CLASS
#define CPL_CLASS CPL_CLASS_INT
#include "cpl_image_basic_body.h"
#undef CPL_CLASS
      default:
        cpl_ensure_code(0, CPL_ERROR_TYPE_MISMATCH);
    }
    return CPL_ERROR_NONE ;
}
#undef CPL_OPERATOR

#define CPL_OPERATOR CPL_IMAGE_EXPASSIGN
/*----------------------------------------------------------------------------*/
/**
  @brief    Compute the elementwise exponential of the image.
  @param    image   Image to be modified in place.
  @param    base    Base of the exponential.
  @return   CPL_ERROR_NONE or the relevant the #_cpl_error_code_ on error
 
  Modifies the image by computing the base-scalar exponential of each of its
  pixels.

  Images can be CPL_TYPE_INT, CPL_TYPE_FLOAT or CPL_TYPE_DOUBLE.

  If the base is zero all pixels must be positive and if the base is
  negative all pixels must be integer, otherwise a cpl_error_code is
  returned and the image is unmodified.

  Possible #_cpl_error_code_ set in this function:
  - CPL_ERROR_NULL_INPUT if (one of) the input pointer(s) is NULL
  - CPL_ERROR_ILLEGAL_INPUT if the requirements on the image are not
    respected
  - CPL_ERROR_TYPE_MISMATCH if the passed image type is not supported
  - CPL_ERROR_DIVISION_BY_ZERO a division by 0 occurs
 */
/*----------------------------------------------------------------------------*/
cpl_error_code cpl_image_exponential(
        cpl_image   *   image,
        double          base)
{
    const double scalar = base;
    int          i ;

    /* Error handling   */
    cpl_ensure_code(image  != NULL, CPL_ERROR_NULL_INPUT);

    assert( image->pixels );
    
    if (base == 0) {
        unsigned char ok = 1;
#undef CPL_OPERATION
#define CPL_OPERATION CPL_IMAGE_CHK_POSITIVE
        /* Switch on image type */
        switch (image->type) {
#define CPL_CLASS CPL_CLASS_DOUBLE
#include "cpl_image_basic_body.h"
#undef CPL_CLASS
#define CPL_CLASS CPL_CLASS_FLOAT
#include "cpl_image_basic_body.h"
#undef CPL_CLASS
#define CPL_CLASS CPL_CLASS_INT
#include "cpl_image_basic_body.h"
#undef CPL_CLASS
          default:
              cpl_ensure_code(0, CPL_ERROR_TYPE_MISMATCH);
        }
#undef CPL_OPERATION
        cpl_ensure_code(ok, CPL_ERROR_DIVISION_BY_ZERO);
    } else if (base < 0) {
        unsigned char ok = 1;
#define CPL_OPERATION CPL_IMAGE_CHK_INTEGER
        /* Switch on image type */
        switch (image->type) {
#define CPL_CLASS CPL_CLASS_DOUBLE
#include "cpl_image_basic_body.h"
#undef CPL_CLASS
#define CPL_CLASS CPL_CLASS_FLOAT
#include "cpl_image_basic_body.h"
#undef CPL_CLASS
          default:
              cpl_ensure_code(0, CPL_ERROR_TYPE_MISMATCH);
        }
#undef CPL_OPERATION
#define CPL_OPERATION CPL_IMAGE_BASIC_OP_SCALAR
        cpl_ensure_code(ok, CPL_ERROR_ILLEGAL_INPUT);
    }
    
    /* Switch on image type */
    switch (image->type) {
#define CPL_CLASS CPL_CLASS_DOUBLE
#include "cpl_image_basic_body.h"
#undef CPL_CLASS
#define CPL_CLASS CPL_CLASS_FLOAT
#include "cpl_image_basic_body.h"
#undef CPL_CLASS
#define CPL_CLASS CPL_CLASS_INT
#include "cpl_image_basic_body.h"
#undef CPL_CLASS
      default:
        cpl_ensure_code(0, CPL_ERROR_TYPE_MISMATCH);
    }
    return CPL_ERROR_NONE ;
}
#undef CPL_OPERATOR
#undef CPL_OPERATION

/*----------------------------------------------------------------------------*/
/**
  @brief    Compute the elementwise power of the image.
  @param    image    Image to be modified in place.
  @param    exponent Scalar exponent.
  @return   CPL_ERROR_NONE or the relevant the #_cpl_error_code_ on error
 
  Modifies the image by lifting each of its pixels to exponent.

  Images can be CPL_TYPE_INT, CPL_TYPE_FLOAT or CPL_TYPE_DOUBLE.

  If the exponent is non-positive all pixels must be non-zero and if
  the exponent is non-integer all pixels must be non-negative,
  otherwise a cpl_error_code is returned and the image is unmodified.

  Possible #_cpl_error_code_ set in this function:
  - CPL_ERROR_NULL_INPUT if (one of) the input pointer(s) is NULL
  - CPL_ERROR_ILLEGAL_INPUT if the requirements on the image are not
    respected
  - CPL_ERROR_TYPE_MISMATCH if the passed image type is not supported
  - CPL_ERROR_DIVISION_BY_ZERO a division by 0 occurs
 */
/*----------------------------------------------------------------------------*/
cpl_error_code cpl_image_power(
        cpl_image   *   image,
        double          exponent)
{
    const double scalar = exponent;
    int          i ;

    /* Error handling   */
    cpl_ensure_code(image  != NULL, CPL_ERROR_NULL_INPUT);

    if (exponent <= 0) {
        unsigned char ok = 1;
#define CPL_OPERATION CPL_IMAGE_CHK_NON_ZERO
        /* Switch on image type */
        switch (image->type) {
#define CPL_CLASS CPL_CLASS_DOUBLE
#include "cpl_image_basic_body.h"
#undef CPL_CLASS
#define CPL_CLASS CPL_CLASS_FLOAT
#include "cpl_image_basic_body.h"
#undef CPL_CLASS
#define CPL_CLASS CPL_CLASS_INT
#include "cpl_image_basic_body.h"
#undef CPL_CLASS
          default:
              cpl_ensure_code(0, CPL_ERROR_TYPE_MISMATCH);
        }
#undef CPL_OPERATION
        cpl_ensure_code(ok, CPL_ERROR_DIVISION_BY_ZERO);
    } else if (exponent != ceil(exponent)) {
        unsigned char ok = 1;
#define CPL_OPERATION CPL_IMAGE_CHK_NON_NEGATIVE
        /* Switch on image type */
        switch (image->type) {
#define CPL_CLASS CPL_CLASS_DOUBLE
#include "cpl_image_basic_body.h"
#undef CPL_CLASS
#define CPL_CLASS CPL_CLASS_FLOAT
#include "cpl_image_basic_body.h"
#undef CPL_CLASS
          default:
              cpl_ensure_code(0, CPL_ERROR_TYPE_MISMATCH);
        }
#undef CPL_OPERATION
        cpl_ensure_code(ok, CPL_ERROR_ILLEGAL_INPUT);
    }
  
    /* Special case with exponent == 0.5 -> use sqrt() for efficiency */
    if (scalar == 0.5) {
#define CPL_OPERATION CPL_IMAGE_BASIC_SQRT
        /* Switch on image type */
        switch (image->type) {
#define CPL_CLASS CPL_CLASS_DOUBLE
#include "cpl_image_basic_body.h"
#undef CPL_CLASS
#define CPL_CLASS CPL_CLASS_FLOAT
#include "cpl_image_basic_body.h"
#undef CPL_CLASS
#define CPL_CLASS CPL_CLASS_INT
#include "cpl_image_basic_body.h"
#undef CPL_CLASS
          default:
            cpl_ensure_code(0, CPL_ERROR_TYPE_MISMATCH);
        }
#undef CPL_OPERATION
    } else {
        /* General case */
#define CPL_OPERATION CPL_IMAGE_BASIC_OP_SCALAR
#define CPL_OPERATOR CPL_IMAGE_POWASSIGN
        /* Switch on image type */
        switch (image->type) {
#define CPL_CLASS CPL_CLASS_DOUBLE
#include "cpl_image_basic_body.h"
#undef CPL_CLASS
#define CPL_CLASS CPL_CLASS_FLOAT
#include "cpl_image_basic_body.h"
#undef CPL_CLASS
#define CPL_CLASS CPL_CLASS_INT
#include "cpl_image_basic_body.h"
#undef CPL_CLASS
          default:
            cpl_ensure_code(0, CPL_ERROR_TYPE_MISMATCH);
        }
#undef CPL_OPERATOR
#undef CPL_OPERATION
    }
    return CPL_ERROR_NONE ;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Normalise pixels in an image.
  @param    image    Image operand.
  @param    mode     Normalisation mode.
  @return   CPL_ERROR_NONE, or the relevant #_cpl_error_code_ on error.
 
  Normalises an image according to a given criterion.
 
  Possible normalisations are:
  - CPL_NORM_SCALE sets the pixel interval to [0,1].
  - CPL_NORM_MEAN sets the mean value to 1.
  - CPL_NORM_FLUX sets the flux to 1.
  - CPL_NORM_ABSFLUX sets the absolute flux to 1.

  Possible #_cpl_error_code_ set in this function:
  - CPL_ERROR_NULL_INPUT if (one of) the input pointer(s) is NULL
 */
/*----------------------------------------------------------------------------*/
cpl_error_code cpl_image_normalise(
        cpl_image * image, 
        cpl_norm    mode)
{
    double scale;


    cpl_ensure_code(image, CPL_ERROR_NULL_INPUT);

    switch (mode) {
        case CPL_NORM_SCALE: {
            cpl_stats * stats =
                cpl_stats_new_from_image(image, CPL_STATS_MIN | CPL_STATS_MAX);

            cpl_ensure_code( stats != NULL, cpl_error_get_code());
            scale = cpl_stats_get_max(stats) - cpl_stats_get_min(stats) ;
            if (scale > 0 &&
                cpl_image_subtract_scalar(image, cpl_stats_get_min(stats))) {
                cpl_stats_delete(stats);
                cpl_ensure_code(0, cpl_error_get_code()) ;
            }
            cpl_stats_delete(stats);
            break ;
        }
        case CPL_NORM_MEAN: {
            scale = cpl_image_get_mean(image);
            break ;
	}
        case CPL_NORM_FLUX: {
            scale = cpl_image_get_flux(image);
            break ;
	}
        case CPL_NORM_ABSFLUX: {
            scale = cpl_image_get_absflux(image);
            break ;
	}
        default:
            /* This case can only be reached if cpl_norm is extended in error */
            assert( 0 ) ;
    }

    cpl_ensure_code( !cpl_image_divide_scalar(image, scale),
                     cpl_error_get_code());

    return CPL_ERROR_NONE ;
}


/*----------------------------------------------------------------------------*/
/**
  @brief    Create a new normalised image from an existing image.
  @param    image_in    Image operand.
  @param    mode        Normalisation mode.
  @return   1 newly allocated image or NULL on error
  @see      cpl_image_normalise 

  Stores the result in a newly allocated image and returns it.
  The returned image must be deallocated using cpl_image_delete().
 
  - CPL_ERROR_NULL_INPUT if (one of) the input pointer(s) is NULL
 */
/*----------------------------------------------------------------------------*/
cpl_image * cpl_image_normalise_create(
        const cpl_image *   image_in, 
        cpl_norm            mode)
{
    cpl_image  *   image_out;


    cpl_ensure(image_in, CPL_ERROR_NULL_INPUT, NULL);

    image_out = cpl_image_duplicate(image_in);

    if (cpl_image_normalise(image_out, mode)) {
        cpl_image_delete(image_out);
        cpl_ensure(0, cpl_error_get_code(), NULL);
    }

    return image_out ;
}


/*----------------------------------------------------------------------------*/
/**
  @brief    Create a new image by elementwise addition of a scalar to an image
  @param    image   Image to add
  @param    addend  Number to add
  @return   1 newly allocated image or NULL in case of an error
  @see      cpl_image_add_scalar

  Creates a new image, being the result of the operation, and returns it to
  the caller. The returned image must be deallocated using cpl_image_delete().
  The function supports images with different types among CPL_TYPE_INT, 
  CPL_TYPE_FLOAT and CPL_TYPE_DOUBLE. The type of the created image is that of
  the passed image.
 */
/*----------------------------------------------------------------------------*/
cpl_image * cpl_image_add_scalar_create(
        const cpl_image   *   image,
        double                addend)
{
    cpl_image  * result =  cpl_image_duplicate(image);

    cpl_ensure(result, cpl_error_get_code(), NULL);

    if (cpl_image_add_scalar(result, addend)) {
        cpl_image_delete(result);
        cpl_ensure(0, cpl_error_get_code(), NULL);
    }
    return result;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Create an image by elementwise subtraction of a scalar from an image
  @param    image       Image to be subtracted from
  @param    subtrahend  Number to subtract
  @return   1 newly allocated image or NULL in case of an error
  @see      cpl_image_subtract_scalar
  @see      cpl_image_add_scalar_create
 */
/*----------------------------------------------------------------------------*/
cpl_image * cpl_image_subtract_scalar_create(
        const cpl_image   *   image,
        double                subtrahend)
{
    cpl_image  * result =  cpl_image_duplicate(image);


    cpl_ensure(result, cpl_error_get_code(), NULL);

    if (cpl_image_subtract_scalar(result, subtrahend)) {
        cpl_image_delete(result);
        cpl_ensure(0, cpl_error_get_code(), NULL);
    }

    return result;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Create a new image by multiplication of a scalar and an image
  @param    image   Image to be multiplied
  @param    factor  Number to multiply with
  @return   1 newly allocated image or NULL in case of an error
  @see      cpl_image_multiply_scalar
  @see      cpl_image_add_scalar_create
 */
/*----------------------------------------------------------------------------*/
cpl_image * cpl_image_multiply_scalar_create(
        const cpl_image   *   image,
        double                factor)
{
    cpl_image  * result =  cpl_image_duplicate(image);


    cpl_ensure(result, cpl_error_get_code(), NULL);

    if (cpl_image_multiply_scalar(result, factor)) {
        cpl_image_delete(result);
        cpl_ensure(0, cpl_error_get_code(), NULL);
    }

    return result;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Create a new image by elementwise division of an image with a scalar
  @param    image    Image to divide
  @param    divisor  Non-zero number to divide with
  @return   1 newly allocated image or NULL in case of an error
  @see      cpl_image_divide_scalar
  @see      cpl_image_add_scalar_create
 */
/*----------------------------------------------------------------------------*/
cpl_image * cpl_image_divide_scalar_create(
        const cpl_image   *   image,
        double                divisor)
{
    cpl_image  * result =  cpl_image_duplicate(image);


    cpl_ensure(result, cpl_error_get_code(), NULL);

    if (cpl_image_divide_scalar(result, divisor)) {
        cpl_image_delete(result);
        cpl_ensure(0, cpl_error_get_code(), NULL);
    }

    return result;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Create a new image by taking the elementwise logarithm of an image
  @param    image   Image to take logarithm of
  @param    base    Base of the logarithm.
  @return   1 newly allocated image or NULL in case of an error
  @see      cpl_image_logarithm
  @see      cpl_image_add_scalar_create
 */
/*----------------------------------------------------------------------------*/
cpl_image * cpl_image_logarithm_create(
        const cpl_image   *   image,
        double          base)
{
    cpl_image  * result =  cpl_image_duplicate(image);

    cpl_ensure(result, cpl_error_get_code(), NULL);

    if (cpl_image_logarithm(result, base)) {
        cpl_image_delete(result);
        cpl_ensure(0, cpl_error_get_code(), NULL);
    }

    return result;
}
        
/*----------------------------------------------------------------------------*/
/**
  @brief    Create a new image by elementwise exponentiation of an image
  @param    image   Image to exponentiate
  @param    base    Base of the exponential
  @return   1 newly allocated image or NULL in case of an error
  @see      cpl_image_logarithm
  @see      cpl_image_add_scalar_create
 */
/*----------------------------------------------------------------------------*/
cpl_image * cpl_image_exponential_create(
        const cpl_image   *   image,
        double          base)
{
    cpl_image  * result =  cpl_image_duplicate(image);

    cpl_ensure(result, cpl_error_get_code(), NULL);

    if (cpl_image_exponential(result, base)) {
        cpl_image_delete(result);
        cpl_ensure(0, cpl_error_get_code(), NULL);
    }

    return result;
}
        
/*----------------------------------------------------------------------------*/
/**
  @brief    Create a new image by elementwise raising of an image to a power
  @param    image    Image to raise to a power
  @param    exponent scalar exponent
  @return   1 newly allocated image or NULL in case of an error
  @see      cpl_image_exponent
  @see      cpl_image_add_scalar_create
 */
/*----------------------------------------------------------------------------*/
cpl_image * cpl_image_power_create(
        const cpl_image   *   image,
        double          exponent)
{
    cpl_image  * result =  cpl_image_duplicate(image);

    cpl_ensure(result, cpl_error_get_code(), NULL);

    if (cpl_image_power(result, exponent)) {
        cpl_image_delete(result);
        cpl_ensure(0, cpl_error_get_code(), NULL);
    }

    return result;
}
        
#define CPL_OPERATION CPL_IMAGE_BASIC_THRESHOLD
/*----------------------------------------------------------------------------*/
/**
  @brief    Threshold an image to a given interval.
  @param    image_in            Image to threshold.
  @param    lo_cut              Lower bound.
  @param    hi_cut              Higher bound.
  @param    assign_lo_cut_in    Value to assign to pixels below low bound.
  @param    assign_hi_cut_in    Value to assign to pixels above high bound.
  @return   the #_cpl_error_code_ or CPL_ERROR_NONE
  
  Pixels outside of the provided interval are assigned the given values. To 
  specify no threshold on lower bound, set lo_cut to CPL_PIXEL_MINVAL. 
  To specify no threshold on higher bound, set hi_cut to CPL_PIXEL_MAXVAL.
  Images can be CPL_TYPE_INT, CPL_TYPE_FLOAT or CPL_TYPE_DOUBLE.
  lo_cut must be small than or equal to hi_cut.

  Possible #_cpl_error_code_ set in this function:
  - CPL_ERROR_NULL_INPUT if (one of) the input pointer(s) is NULL
  - CPL_ERROR_TYPE_MISMATCH if the passed image type is not supported
  - CPL_ERROR_ILLEGAL_INPUT if lo_cut is greater than hi_cut
 */
/*----------------------------------------------------------------------------*/
cpl_error_code cpl_image_threshold(
        cpl_image *   image_in,
        double              lo_cut,
        double              hi_cut,
        double              assign_lo_cut_in,
        double              assign_hi_cut_in)
{
    const double    assign_lo_cut = assign_lo_cut_in == (CPL_PIXEL_MINVAL)
                                  ? lo_cut : assign_lo_cut_in;
    const double    assign_hi_cut = assign_hi_cut_in == (CPL_PIXEL_MAXVAL)
                                  ? hi_cut : assign_hi_cut_in;
    int             i ;

    cpl_ensure_code(image_in != NULL, CPL_ERROR_NULL_INPUT);
    cpl_ensure_code(lo_cut <= hi_cut, CPL_ERROR_ILLEGAL_INPUT);

    /* Switch on image type */
    switch (image_in->type) {
#define CPL_CLASS CPL_CLASS_DOUBLE
#include "cpl_image_basic_body.h"
#undef CPL_CLASS
#define CPL_CLASS CPL_CLASS_FLOAT
#include "cpl_image_basic_body.h"
#undef CPL_CLASS
#define CPL_CLASS CPL_CLASS_INT
#include "cpl_image_basic_body.h"
#undef CPL_CLASS
        default:
          cpl_ensure_code(0, CPL_ERROR_TYPE_MISMATCH);
    }
    return CPL_ERROR_NONE ;
}
#undef CPL_OPERATION

#define CPL_OPERATION CPL_IMAGE_BASIC_ABS
/*----------------------------------------------------------------------------*/
/**
  @brief    Take the absolute value of an image.
  @param    image    Image to be modified in place
  @return   CPL_ERROR_NONE or the relevant the #_cpl_error_code_ on error
  
  Set each pixel to its absolute value.

  Possible #_cpl_error_code_ set in this function:
  - CPL_ERROR_NULL_INPUT if (one of) the input pointer(s) is NULL
  - CPL_ERROR_TYPE_MISMATCH if the passed image type is not supported
 */
/*----------------------------------------------------------------------------*/
cpl_error_code cpl_image_abs(cpl_image * image)
{
    int                i ;

    /* Check entries */
    cpl_ensure_code(image, CPL_ERROR_NULL_INPUT);

    /* Switch on image type */
    switch (image->type) {
#define CPL_CLASS CPL_CLASS_DOUBLE
#include "cpl_image_basic_body.h"
#undef CPL_CLASS
#define CPL_CLASS CPL_CLASS_FLOAT
#include "cpl_image_basic_body.h"
#undef CPL_CLASS
#define CPL_CLASS CPL_CLASS_INT
#include "cpl_image_basic_body.h"
#undef CPL_CLASS
        default:
            cpl_ensure_code(0, CPL_ERROR_TYPE_MISMATCH);
    }

    return CPL_ERROR_NONE;
}

#undef CPL_OPERATION

/*----------------------------------------------------------------------------*/
/**
  @brief    Take the absolute value of an image.
  @param    image_in    Image operand.
  @return   1 newly allocated image or NULL in error case
  @see      cpl_image_abs
  
  For each pixel, out = abs(in). The returned image must be deallocated using
  cpl_image_delete().
 */
/*----------------------------------------------------------------------------*/
cpl_image * cpl_image_abs_create(const cpl_image * image_in)
{

    cpl_image  * result = cpl_image_duplicate(image_in);

    cpl_ensure(result, cpl_error_get_code(), NULL);

    if (cpl_image_abs(result)) {
        cpl_image_delete(result);
        cpl_ensure(0, cpl_error_get_code(), NULL);
    }

    return result;
}

#define CPL_OPERATION CPL_IMAGE_BASIC_AVERAGE
/*----------------------------------------------------------------------------*/
/**
  @brief    Build the average of two images.
  @param    image_1     First image operand.
  @param    image_2     Second image operand.
  @return   1 newly allocated image or NULL in error case
 
  Builds the average of two images and returns a newly allocated image,
  to be deallocated using cpl_image_delete(). The average is arithmetic, i.e.
  outpix=(pix1+pix2)/2
  Images can be CPL_TYPE_INT, CPL_TYPE_FLOAT or CPL_TYPE_DOUBLE.

  Possible #_cpl_error_code_ set in this function:
  - CPL_ERROR_NULL_INPUT if (one of) the input pointer(s) is NULL
  - CPL_ERROR_TYPE_MISMATCH if the passed image type is not supported
 */
/*----------------------------------------------------------------------------*/
cpl_image * cpl_image_average_create(
        const cpl_image *   image_1,
        const cpl_image *   image_2)
{   
    cpl_image   *   image_out ;
    int         *   pii2;
    float       *   pfi2;
    double      *   pdi2;
    int             i ;

    /* Check entries   */
    cpl_ensure(image_1 && image_2, CPL_ERROR_NULL_INPUT, NULL);
   
    /* Switch on first passed image type */
    switch (image_1->type) {
#define CPL_CLASS CPL_CLASS_DOUBLE
#include "cpl_image_basic_body.h"
#undef CPL_CLASS
#define CPL_CLASS CPL_CLASS_FLOAT
#include "cpl_image_basic_body.h"
#undef CPL_CLASS
#define CPL_CLASS CPL_CLASS_INT
#include "cpl_image_basic_body.h"
#undef CPL_CLASS
        default:
          cpl_ensure(0, CPL_ERROR_TYPE_MISMATCH, 
                  NULL) ;
    }
    
    /* Handle bad pixels map */
    if (image_1->bpm == NULL && image_2->bpm == NULL) {
        image_out->bpm = NULL ;
    } else if (image_1->bpm == NULL) {
        image_out->bpm = cpl_mask_duplicate(image_2->bpm) ;
    } else if (image_2->bpm == NULL) {
        image_out->bpm = cpl_mask_duplicate(image_1->bpm) ;
    } else {
        image_out->bpm = cpl_mask_duplicate(image_1->bpm) ;
        cpl_mask_or(image_out->bpm, image_2->bpm) ;
    }

    return image_out ;
}
#undef CPL_OPERATION


/*----------------------------------------------------------------------------*/
/**
  @brief    Collapse an image region along its rows or columns.
  @param    in          Image to collapse.
  @param    llx         lower left x coord.
  @param    lly         lower left y coord
  @param    urx         upper right x coord
  @param    ury         upper right y coord
  @param    direction   Collapsing direction.
  @return   a newly allocated image or NULL in error case
  @see      cpl_image_collapse_create()
  
  llx, lly, urx, ury are the image region coordinates in FITS convention.
  Those specified bounds are included in the collapsed region.
  Images can be CPL_TYPE_INT, CPL_TYPE_FLOAT or CPL_TYPE_DOUBLE.
  The returned image must be deallocated using cpl_image_delete().

  Possible #_cpl_error_code_ set in this function:
  - CPL_ERROR_NULL_INPUT if (one of) the input pointer(s) is NULL
  - CPL_ERROR_ILLEGAL_INPUT if the specified window is not valid
 */
/*----------------------------------------------------------------------------*/
cpl_image * cpl_image_collapse_window_create(
        const cpl_image *   in,
        int                 llx,
        int                 lly,
        int                 urx,
        int                 ury,
        int                 direction)
{
    cpl_image   *   extracted ;
    cpl_image   *   collapsed ;

    cpl_ensure(in, CPL_ERROR_NULL_INPUT, NULL);

    /* First extract the image to collapse */
    cpl_ensure((extracted = cpl_image_extract(in, llx, lly, urx, ury)),
               CPL_ERROR_ILLEGAL_INPUT, NULL) ;

    /* Collapse now the extracted image */
    if ((collapsed = cpl_image_collapse_create(extracted, direction)) == NULL) {
        cpl_image_delete(extracted) ;
        cpl_ensure(0, CPL_ERROR_ILLEGAL_INPUT, NULL) ;
    }
    cpl_image_delete(extracted) ;

    return collapsed ;
}

#define CPL_OPERATION CPL_IMAGE_BASIC_COLLAPSE
/*----------------------------------------------------------------------------*/
/**
  @brief    Collapse an image along its rows or columns.
  @param    inimage     Image to collapse.
  @param    direction   Collapsing direction.
  @return   1 newly allocated image having 1 row or 1 column or NULL for error
 
  Collapsing an image means building up a 1d signal by adding up all pixels
  on the same row or column.
 
  @verbatim
  Collapse along y:
 
  p7  p8  p9     Input image is a 3x3 image containing 9 pixels.
  p4  p5  p6     The output is an image containing one row with
  p1  p2  p3     3 pixels A, B, C, where:
  ----------
 
  A   B   C      A = p1+p4+p7
                 B = p2+p5+p8
                 C = p3+p6+p9

  If p7 is a bad pixel, A = (p1+p4)*3/2.
  If p1, p4, p7 are bad, A is flagged as bad.
  @endverbatim
 
  Provide the collapsing direction as an int. Give 0 to collapse along y
  (sum of rows) and get an image with a single row in output, or give 1
  to collapse along x (sym of columns) to get an image with a single
  column in output.
  Only the good pixels are collapsed.
  Images can be CPL_TYPE_INT, CPL_TYPE_FLOAT or CPL_TYPE_DOUBLE.
  The returned image must be deallocated using cpl_image_delete().

  Possible #_cpl_error_code_ set in this function:
  - CPL_ERROR_NULL_INPUT if (one of) the input pointer(s) is NULL
  - CPL_ERROR_TYPE_MISMATCH if the passed image type is not supported
 */
/*----------------------------------------------------------------------------*/
cpl_image * cpl_image_collapse_create(
        const cpl_image *   inimage,
        int                 direction)
{
    cpl_image   *   image1d ;
    double          mean ;
    int             i ;

    /* Test entries */
    cpl_ensure(inimage, CPL_ERROR_NULL_INPUT, NULL);
   
    /* Switch on image type */
    switch (inimage->type) {
#define CPL_CLASS CPL_CLASS_DOUBLE
#include "cpl_image_basic_body.h"
#undef CPL_CLASS
#define CPL_CLASS CPL_CLASS_FLOAT
#include "cpl_image_basic_body.h"
#undef CPL_CLASS
#define CPL_CLASS CPL_CLASS_INT
#include "cpl_image_basic_body.h"
#undef CPL_CLASS
        default:
          cpl_ensure(0, CPL_ERROR_TYPE_MISMATCH, NULL) ;
    }
    
    /* Handle the bad pixels map */
    if (inimage->bpm == NULL) image1d->bpm = NULL ;
    else image1d->bpm = cpl_mask_collapse_create(inimage->bpm, direction) ;

    /* Return */
    return image1d ;
}
#undef CPL_OPERATION

#define CPL_OPERATION CPL_IMAGE_BASIC_COLLAPSE_MEDIAN
/*----------------------------------------------------------------------------*/
/**
  @brief    Collapse an image along its rows and columns, with filtering.
  @param    in          Input image.
  @param    direction   Collapsing direction.
  @param    discard_lo  Low rejection parameter.
  @param    discard_hi  High rejection parameter.
  @return   1 newly allocated image having 1 row or 1 column or NULL for error
  @see      cpl_image_collapse_create()
  
  Collapsing is done as for cpl_image_collapse_create(). The difference is 
  that pixels are not just summed along rows or columns but sorted,
  the highest and lowest pixel values are removed, and the median is 
  computed on the remaining values.
  All pixels are collapsed, even the bad pixels. The bad pixel map of
  the collapsed image is empty whatever the input is, the bad pixel maps
  are not used in this function.
  Images can be CPL_TYPE_INT, CPL_TYPE_FLOAT or CPL_TYPE_DOUBLE.
  The returned image must be deallocated using cpl_image_delete().

  Possible #_cpl_error_code_ set in this function:
  - CPL_ERROR_NULL_INPUT if (one of) the input pointer(s) is NULL
  - CPL_ERROR_ILLEGAL_INPUT if the sum of discarded pixels is bigger
    than the image size in the collapsing direction
  - CPL_ERROR_TYPE_MISMATCH if the passed image type is not supported
 */
/*----------------------------------------------------------------------------*/
cpl_image * cpl_image_collapse_median_create(
        const cpl_image *   in,
        int                 direction,
        int                 discard_lo,
        int                 discard_hi)
{
    cpl_image   *   collapsed ;
    double      *   line ;
    int             width ;
    int             i, j ;

    /* Test entries */
    cpl_ensure(in, CPL_ERROR_NULL_INPUT, NULL);
    cpl_ensure(!((direction == 0) && (discard_lo+discard_hi>=in->ny)),
               CPL_ERROR_ILLEGAL_INPUT, NULL) ;
    cpl_ensure(!((direction == 1) && (discard_lo+discard_hi>=in->nx)),
               CPL_ERROR_ILLEGAL_INPUT, NULL) ;
    
    /* Switch on image type */
    switch (in->type) {
#define CPL_CLASS CPL_CLASS_DOUBLE
#include "cpl_image_basic_body.h"
#undef CPL_CLASS
#define CPL_CLASS CPL_CLASS_FLOAT
#include "cpl_image_basic_body.h"
#undef CPL_CLASS
#define CPL_CLASS CPL_CLASS_INT
#include "cpl_image_basic_body.h"
#undef CPL_CLASS

        default:
          cpl_ensure(0, CPL_ERROR_TYPE_MISMATCH, NULL) ;
    } 
    collapsed->bpm = NULL ;
    
    /* Return */
    return collapsed ;
}
#undef CPL_OPERATION

#define CPL_OPERATION CPL_IMAGE_BASIC_EXTRACT
/*----------------------------------------------------------------------------*/
/**
  @brief    Extract a rectangular zone from an image into another image.
  @param    image_in    Input image
  @param    loleft_x    Lower left X coordinate
  @param    loleft_y    Lower left Y coordinate
  @param    upright_x   Upper right X coordinate
  @param    upright_y   Upper right Y coordinate
  @return   1 newly allocated image or NULL in error case
 
  The input coordinates define the extracted region by giving the coordinates 
  of the lower left and upper right corners (inclusive).
 
  Coordinates must be provided in the FITS convention: lower left
  corner of the image is at (1,1), x increasing from left to right,
  y increasing from bottom to top.
  Images can be CPL_TYPE_INT, CPL_TYPE_FLOAT or CPL_TYPE_DOUBLE.
  The returned image must be deallocated using cpl_image_delete().

  Possible #_cpl_error_code_ set in this function:
  - CPL_ERROR_NULL_INPUT if (one of) the input pointer(s) is NULL
  - CPL_ERROR_ILLEGAL_INPUT if the window coordinates are not valid
  - CPL_ERROR_TYPE_MISMATCH if the passed image type is not supported
 */
/*----------------------------------------------------------------------------*/
cpl_image * cpl_image_extract(
        const cpl_image *   image_in,
        int                 loleft_x,
        int                 loleft_y,
        int                 upright_x,
        int                 upright_y)
{
    cpl_image   *   slit_img ;
    int             outlx,
                    outly ;
    int             i, j ;

    cpl_ensure(image_in, CPL_ERROR_NULL_INPUT, NULL);
    cpl_ensure(!((loleft_x<1) || (loleft_x>image_in->nx) ||
        (loleft_y<1) || (loleft_y>image_in->ny) ||
        (upright_x<1) || (upright_x>image_in->nx) ||
        (upright_y<1) || (upright_y>image_in->ny) ||
        (loleft_x>upright_x) || (loleft_y>upright_y)),
               CPL_ERROR_ILLEGAL_INPUT, NULL) ;

    outlx = upright_x - loleft_x + 1 ;
    outly = upright_y - loleft_y + 1 ;

    /* Switch on image type */
    switch (image_in->type) {
#define CPL_CLASS CPL_CLASS_DOUBLE
#include "cpl_image_basic_body.h"
#undef CPL_CLASS
#define CPL_CLASS CPL_CLASS_FLOAT
#include "cpl_image_basic_body.h"
#undef CPL_CLASS
#define CPL_CLASS CPL_CLASS_INT
#include "cpl_image_basic_body.h"
#undef CPL_CLASS
        default:
            cpl_ensure(0, CPL_ERROR_TYPE_MISMATCH, NULL) ;
    }
    
    /* Bad pixels handling */
    if (image_in->bpm == NULL) slit_img->bpm = NULL ;
    else slit_img->bpm = cpl_mask_extract(image_in->bpm, loleft_x,
            loleft_y, upright_x, upright_y);

    return slit_img ;
}
#undef CPL_OPERATION

#define CPL_OPERATION CPL_IMAGE_BASIC_EXTRACTROW
/*----------------------------------------------------------------------------*/
/**
  @brief    Extract a row from an image
  @param    image_in    Input image
  @param    pos         Position of the row (1 for the bottom one)
  @return   1 newly allocated cpl_vector or NULL in error case
 
  Images can be CPL_TYPE_INT, CPL_TYPE_FLOAT or CPL_TYPE_DOUBLE.
  The returned vector must be deallocated using cpl_vector_delete().

  The bad pixels map is not taken into account in this function.
  
  Possible #_cpl_error_code_ set in this function:
  - CPL_ERROR_NULL_INPUT if (one of) the input pointer(s) is NULL
  - CPL_ERROR_ILLEGAL_INPUT if pos is not valid
  - CPL_ERROR_TYPE_MISMATCH if the passed image type is not supported
 */
/*----------------------------------------------------------------------------*/
cpl_vector * cpl_vector_new_from_image_row(
        const cpl_image *   image_in,
        int                 pos)
{
    cpl_vector  *   out ;
    double      *   out_data ;
    int             i ;

    /* Test entries */
    cpl_ensure(image_in, CPL_ERROR_NULL_INPUT, NULL);
    cpl_ensure(pos>=1 && pos<=image_in->ny, CPL_ERROR_ILLEGAL_INPUT,NULL);
    
    /* Allocate output vector */
    out = cpl_vector_new(image_in->nx) ;
    out_data = cpl_vector_get_data(out) ;
    
    /* Switch on image type */
    switch (image_in->type) {
#define CPL_CLASS CPL_CLASS_DOUBLE
#include "cpl_image_basic_body.h"
#undef CPL_CLASS
#define CPL_CLASS CPL_CLASS_FLOAT
#include "cpl_image_basic_body.h"
#undef CPL_CLASS
#define CPL_CLASS CPL_CLASS_INT
#include "cpl_image_basic_body.h"
#undef CPL_CLASS
        default:
            cpl_vector_delete(out) ;
            cpl_ensure(0, CPL_ERROR_TYPE_MISMATCH, NULL) ;
    }
    
    /* Return */
    return out ;
}
#undef CPL_OPERATION

#define CPL_OPERATION CPL_IMAGE_BASIC_EXTRACTCOL
/*----------------------------------------------------------------------------*/
/**
  @brief    Extract a column from an image
  @param    image_in    Input image
  @param    pos         Position of the column (1 for the left one)
  @return   1 newly allocated cpl_vector or NULL in error case
 
  Images can be CPL_TYPE_INT, CPL_TYPE_FLOAT or CPL_TYPE_DOUBLE.
  The returned vector must be deallocated using cpl_vector_delete().

  The bad pixels map is not taken into account in this function.
  
  Possible #_cpl_error_code_ set in this function:
  - CPL_ERROR_NULL_INPUT if (one of) the input pointer(s) is NULL
  - CPL_ERROR_ILLEGAL_INPUT if pos is not valid
  - CPL_ERROR_TYPE_MISMATCH if the passed image type is not supported
 */
/*----------------------------------------------------------------------------*/
cpl_vector * cpl_vector_new_from_image_column(
        const cpl_image *   image_in,
        int                 pos)
{
    cpl_vector  *   out ;
    double      *   out_data ;
    int             i ;

    /* Check entries */
    cpl_ensure(image_in, CPL_ERROR_NULL_INPUT, NULL);
    cpl_ensure(pos>=1 && pos<=image_in->nx, CPL_ERROR_ILLEGAL_INPUT,NULL);
    
    /* Allocate output vector */
    out = cpl_vector_new(image_in->ny) ;
    out_data = cpl_vector_get_data(out) ;
    
    /* Switch on image type */
    switch (image_in->type) {
#define CPL_CLASS CPL_CLASS_DOUBLE
#include "cpl_image_basic_body.h"
#undef CPL_CLASS
#define CPL_CLASS CPL_CLASS_FLOAT
#include "cpl_image_basic_body.h"
#undef CPL_CLASS
#define CPL_CLASS CPL_CLASS_INT
#include "cpl_image_basic_body.h"
#undef CPL_CLASS
        default:
            cpl_vector_delete(out) ;
            cpl_ensure(0, CPL_ERROR_TYPE_MISMATCH, NULL) ;
    }
    
    /* Return */
    return out ;
}
#undef CPL_OPERATION

#define CPL_OPERATION CPL_IMAGE_BASIC_ROTATE_INT_LOCAL
/*----------------------------------------------------------------------------*/
/**
  @brief    Rotate an image by a multiple of 90 deg clockwise.
  @param    im      the image to rotate.
  @param    rot     the mutiple: -1 is a rotation of 90 deg counterclockwise.
  @return   the #_cpl_error_code_ or CPL_ERROR_NONE

  This function operates locally on the pixel buffer.

  Images can be CPL_TYPE_INT, CPL_TYPE_FLOAT or CPL_TYPE_DOUBLE.

  rot can take one of the following values:
  - +1 to turn 90 degrees clockwise.
  - -1 to turn 90 degrees counterclockwise.
  - +2 to turn 180 degrees.

  Possible #_cpl_error_code_ set in this function:
  - CPL_ERROR_NULL_INPUT if (one of) the input pointer(s) is NULL
  - CPL_ERROR_ILLEGAL_INPUT if rot is different from the allowed values
  - CPL_ERROR_TYPE_MISMATCH if the passed image type is not supported
*/
/*----------------------------------------------------------------------------*/
cpl_error_code cpl_image_turn(
        cpl_image   *   im,
        int             rot)
{
    int             nx, ny ;
    cpl_image   *   tmp_im ;
    int             i, j ;

    /* Check entries */
    cpl_ensure_code(im, CPL_ERROR_NULL_INPUT);
    cpl_ensure_code(rot==1 || rot==-1 || rot==2, CPL_ERROR_ILLEGAL_INPUT);

    /* Initialize */
    nx = cpl_image_get_size_x(im) ;
    ny = cpl_image_get_size_y(im) ;

    /* Duplicate the input image */
    tmp_im = cpl_image_duplicate(im) ;

    /* Switch on the image type */
    switch (im->type) {
#define CPL_CLASS CPL_CLASS_DOUBLE
#include "cpl_image_basic_body.h"
#undef CPL_CLASS
#define CPL_CLASS CPL_CLASS_FLOAT
#include "cpl_image_basic_body.h"
#undef CPL_CLASS
#define CPL_CLASS CPL_CLASS_INT
#include "cpl_image_basic_body.h"
#undef CPL_CLASS
        default:
            cpl_ensure(0, CPL_ERROR_TYPE_MISMATCH,
                    CPL_ERROR_TYPE_MISMATCH) ;
    }
    cpl_image_delete(tmp_im) ;

    /* Rotate the bad pixel map */
    if (im->bpm != NULL) cpl_mask_turn(im->bpm, rot) ;

    return CPL_ERROR_NONE ;
}
#undef CPL_OPERATION

#define CPL_OPERATION CPL_IMAGE_BASIC_SHIFT_INT_LOCAL
/*----------------------------------------------------------------------------*/
/**
  @brief    Shift an image by integer offsets
  @param    im      the image to shift.
  @param    x_shift x shift
  @param    y_shift y shift
  @return   the #_cpl_error_code_ or CPL_ERROR_NONE

  This function operates locally on the pixel buffer.
  The new zones (in the result image) where no new value is computed are set 
  to 0 and flagged as bad pixels.

  Images can be CPL_TYPE_INT, CPL_TYPE_FLOAT or CPL_TYPE_DOUBLE.

  Possible #_cpl_error_code_ set in this function:
  - CPL_ERROR_NULL_INPUT if (one of) the input pointer(s) is NULL
  - CPL_ERROR_ILLEGAL_INPUT if the requested shift is bigger than the
    image size
  - CPL_ERROR_TYPE_MISMATCH if the passed image type is not supported
*/
/*----------------------------------------------------------------------------*/
cpl_error_code cpl_image_shift(
        cpl_image   *   im,
        int             x_shift,
        int             y_shift)
{
    int                 nx, ny ;
    cpl_image       *   tmp_im ;
    cpl_mask        *   bpm ;
    int                 i, j ;

    /* Check entries */
    cpl_ensure(im, CPL_ERROR_NULL_INPUT, CPL_ERROR_NULL_INPUT) ;
    /* Initialize */
    nx = cpl_image_get_size_x(im) ;
    ny = cpl_image_get_size_y(im) ;
    cpl_ensure(nx>fabs(x_shift) && ny>fabs(y_shift), CPL_ERROR_ILLEGAL_INPUT,
             CPL_ERROR_ILLEGAL_INPUT) ;

    /* Duplicate the input image */
    tmp_im = cpl_image_duplicate(im) ;

    /* Switch on the image type */
    switch (im->type) {
#define CPL_CLASS CPL_CLASS_DOUBLE
#include "cpl_image_basic_body.h"
#undef CPL_CLASS
#define CPL_CLASS CPL_CLASS_FLOAT
#include "cpl_image_basic_body.h"
#undef CPL_CLASS
#define CPL_CLASS CPL_CLASS_INT
#include "cpl_image_basic_body.h"
#undef CPL_CLASS
        default:
            cpl_image_delete(tmp_im);
            cpl_ensure_code(0, CPL_ERROR_TYPE_MISMATCH);
    }
    cpl_image_delete(tmp_im) ;

    /* Shift the bad pixel map */
    if (im->bpm == NULL)    bpm = cpl_mask_new(im->nx, im->ny) ;
    else                    bpm = cpl_mask_duplicate(im->bpm) ;
    cpl_mask_shift(bpm, x_shift, y_shift) ;
    cpl_image_reject_from_mask(im, bpm) ;
    cpl_mask_delete(bpm) ;

    return CPL_ERROR_NONE ;
}
#undef CPL_OPERATION

#define CPL_OPERATION CPL_IMAGE_BASIC_COPY
/*----------------------------------------------------------------------------*/
/**
  @brief    Copy one image into another
  @param    im1     the image in which im2 is inserted
  @param    im2     the inserted image
  @param    xpos    the x pixel position in im1 where the lower left pixel of 
                    im2 should go (from 1 to the x size of im1)
  @param    ypos    the y pixel position in im1 where the lower left pixel of 
                    im2 should go (from 1 to the y size of im1)
  @return   CPL_ERROR_NONE or the relevant #_cpl_error_code_ on error.
 
  (xpos, ypos) must be a valid position in im1. If im2 is bigger than the place
  left in im1, the part that falls outside of im1 is simply ignored, an no 
  error is raised.
  The bad pixels are inherited from im2 in the concerned im1 zone.
  
  The two input images must be of the same type, namely one of
  CPL_TYPE_INT, CPL_TYPE_FLOAT, CPL_TYPE_DOUBLE.

  Possible #_cpl_error_code_ set in this function:
  - CPL_ERROR_TYPE_MISMATCH if the passed image type is not supported or
    if the input images are of different types
  - CPL_ERROR_ACCESS_OUT_OF_RANGE if xpos or ypos are outside the
    specified range
*/
/*----------------------------------------------------------------------------*/
cpl_error_code cpl_image_copy(
        cpl_image       *   im1,
        const cpl_image *   im2,
        int                 xpos,
        int                 ypos)
{
    const int       nx1 = cpl_image_get_size_x(im1) ; 
    const int       ny1 = cpl_image_get_size_y(im1) ; 
    const int       nx2 = cpl_image_get_size_x(im2) ; 
    const int       ny2 = cpl_image_get_size_y(im2) ; 
    cpl_mask    *   bpm1 ;
    cpl_mask    *   bpm2 ;
    int             urx, ury ;
    int             im1_pos, im2_pos ;
    int             i, j ;
    
    /* Check entries */
    cpl_ensure_code(nx1 > 0,              cpl_error_get_code());
    cpl_ensure_code(ny1 > 0,              cpl_error_get_code());
    cpl_ensure_code(nx2 > 0,              cpl_error_get_code());
    cpl_ensure_code(ny2 > 0,              cpl_error_get_code());
    cpl_ensure_code(xpos >= 1,           CPL_ERROR_ACCESS_OUT_OF_RANGE);
    cpl_ensure_code(xpos <= nx1,         CPL_ERROR_ACCESS_OUT_OF_RANGE);
    cpl_ensure_code(ypos >= 1,           CPL_ERROR_ACCESS_OUT_OF_RANGE);
    cpl_ensure_code(ypos <= ny1,         CPL_ERROR_ACCESS_OUT_OF_RANGE);
    cpl_ensure_code(im1->type == im2->type, CPL_ERROR_TYPE_MISMATCH);

    /* Define the zone to modify in im1: xpos, ypos, urx, ury */
    if (xpos+nx2-1 >= nx1) urx = nx1;
    else urx = xpos+nx2-1;
    if (ypos+ny2-1 >= ny1) ury = ny1;
    else ury = ypos+ny2-1;

    /* Switch on the image type */
    switch (im1->type) {
#define CPL_CLASS CPL_CLASS_DOUBLE
#include "cpl_image_basic_body.h"
#undef CPL_CLASS
#define CPL_CLASS CPL_CLASS_FLOAT
#include "cpl_image_basic_body.h"
#undef CPL_CLASS
#define CPL_CLASS CPL_CLASS_INT
#include "cpl_image_basic_body.h"
#undef CPL_CLASS
    default:
        cpl_ensure_code(0, CPL_ERROR_TYPE_MISMATCH) ;
    }

    /* Handle the bad pixels */
    if (im1->bpm != NULL || im2->bpm != NULL) {
        if (im1->bpm == NULL)   bpm1 = cpl_mask_new(im1->nx, im1->ny) ;
        else                    bpm1 = cpl_mask_duplicate(im1->bpm) ;
        if (im2->bpm == NULL)   bpm2 = cpl_mask_new(im2->nx, im2->ny) ;
        else                    bpm2 = cpl_mask_duplicate(im2->bpm) ;
        cpl_mask_copy(bpm1, bpm2, xpos, ypos) ;
        cpl_mask_delete(bpm2) ;
        cpl_image_reject_from_mask(im1, bpm1) ;
        cpl_mask_delete(bpm1) ;
    }

    /* Return */
    return CPL_ERROR_NONE ;
}
#undef CPL_OPERATION

#define CPL_OPERATION CPL_IMAGE_BASIC_FLIP_LOCAL
/*----------------------------------------------------------------------------*/
/**
  @brief    Flip an image on a given miror line.
  @param    im      the image to flip.  
  @param    angle   mirror line in polar coord. is theta = (PI/4) * angle 
  @return   the #_cpl_error_code_ or CPL_ERROR_NONE
 
  This function operates locally on the pixel buffer.

  angle can take one of the following values:
  - 0 (theta=0) to flip the image around the horizontal
  - 1 (theta=pi/4) to flip the image around y=x
  - 2 (theta=pi/2) to flip the image around the vertical
  - 3 (theta=3pi/4) to flip the image around y=-x

  Images can be CPL_TYPE_INT, CPL_TYPE_FLOAT or CPL_TYPE_DOUBLE.

  Possible #_cpl_error_code_ set in this function:
  - CPL_ERROR_NULL_INPUT if (one of) the input pointer(s) is NULL
  - CPL_ERROR_ILLEGAL_INPUT if the angle is different from the allowed values
  - CPL_ERROR_TYPE_MISMATCH if the passed image type is not supported
*/
/*----------------------------------------------------------------------------*/
cpl_error_code cpl_image_flip(
        cpl_image   *   im,
        int             angle)
{
    int             nx, ny ;
    cpl_image   *   tmp_im ;
    int             i, j ;
    
    /* Check entries */
    cpl_ensure_code(im, CPL_ERROR_NULL_INPUT) ;
    cpl_ensure_code(angle==0 || angle==1 || angle==2 || angle==3,
            CPL_ERROR_ILLEGAL_INPUT) ;
    
    /* Initialise */
    nx = cpl_image_get_size_x(im) ; 
    ny = cpl_image_get_size_y(im) ; 
   
    /* Duplicate the input image */
    /* FIXME: Avoid this by swapping pairs of pixels */
    tmp_im = cpl_image_duplicate(im) ;
    
    /* Switch on the image type */
    switch (im->type) {
#define CPL_CLASS CPL_CLASS_DOUBLE
#include "cpl_image_basic_body.h"
#undef CPL_CLASS
#define CPL_CLASS CPL_CLASS_FLOAT
#include "cpl_image_basic_body.h"
#undef CPL_CLASS
#define CPL_CLASS CPL_CLASS_INT
#include "cpl_image_basic_body.h"
#undef CPL_CLASS
        default:
            cpl_ensure(0, CPL_ERROR_TYPE_MISMATCH, 
                    CPL_ERROR_TYPE_MISMATCH) ;
    }
    cpl_image_delete(tmp_im) ;

    /* Flip the bad pixel map */
    if (im->bpm != NULL) cpl_mask_flip(im->bpm, angle) ;

    return CPL_ERROR_NONE ;
}
#undef CPL_OPERATION


#define CPL_OPERATION CPL_IMAGE_BASIC_MOVE_PIXELS
/*----------------------------------------------------------------------------*/
/**
  @brief    Reorganize the pixels in an image
  @param    im          the image to reorganize
  @param    nb_cut      the number of cut in x and y
  @param    new_pos     array with the nb_cut^2 new positions
  @return   the #_cpl_error_code_ or CPL_ERROR_NONE

  nb_cut^2 defines in how many tiles the images will be moved. Each tile will
  then be moved to an other place defined in new_pos. 1 will leave the image
  unchanged, 2 is used to move the quadrants, etc..
  new_pos contains nb_cut^2 values between 1 and nb_cut^2.
  The zones positions are counted from the lower left part of the image.
  It is not allowed to move two tiles to the same position (the relation
  between th new tiles positions and the initial position is bijective !).

  Images can be CPL_TYPE_INT, CPL_TYPE_FLOAT or CPL_TYPE_DOUBLE.
  
  The image x and y sizes have to be multiples of nb_cut.

  @verbatim
  Example:

  16   17   18           6    5    4
  13   14   15           3    2    1

  10   11   12   ---->  12   11   10
   7    8    9           9    8    7

   4    5    6          18   17   16
   1    2    3          15   14   13

   image 3x6            cpl_image_move(image, 3, new_pos) ;
                        with new_pos = {9,8,7,6,5,4,3,2,1} ;
   @endverbatim

  The bad pixels are moved accordingly.

  Possible #_cpl_error_code_ set in this function:
  - CPL_ERROR_NULL_INPUT if (one of) the input pointer(s) is NULL
  - CPL_ERROR_ILLEGAL_INPUT if nb_cut is not strictly positive or cannot
    divide one of the image sizes or if the new_pos array specifies to
    move two tiles to the same position.
  - CPL_ERROR_TYPE_MISMATCH if the passed image type is not supported
*/
/*----------------------------------------------------------------------------*/
cpl_error_code cpl_image_move(
        cpl_image   *   im,
        int             nb_cut,
        const int   *   new_pos)
{
    int             test_sum ;
    int             tile_sz_x, tile_sz_y ;
    int             tile_x, tile_y ;
    int             npos, opos ;
    int             i, j, k, l ;

    /* Check entries */
    cpl_ensure_code(im,                   CPL_ERROR_NULL_INPUT) ;
    cpl_ensure_code(new_pos,              CPL_ERROR_NULL_INPUT) ;
    cpl_ensure_code(nb_cut > 0,           CPL_ERROR_ILLEGAL_INPUT) ;
    cpl_ensure_code(im->nx % nb_cut == 0, CPL_ERROR_ILLEGAL_INPUT) ;
    cpl_ensure_code(im->ny % nb_cut == 0, CPL_ERROR_ILLEGAL_INPUT) ;

    /* Test that new_pos takes all values between 1 and nb_cut*nb_cut */
    /* The test here is not strict, but should be sufficient */
    test_sum = 0 ;
    for (i=0 ; i<nb_cut*nb_cut ; i++) test_sum += new_pos[i] ;
    cpl_ensure_code(test_sum==nb_cut*nb_cut*(nb_cut*nb_cut+1)/2,
            CPL_ERROR_ILLEGAL_INPUT) ;

    /* Initialize */
    tile_sz_x = im->nx / nb_cut ;
    tile_sz_y = im->ny / nb_cut ;

    /* Switch on the image type */
    switch (im->type) {
#define CPL_CLASS CPL_CLASS_DOUBLE
#include "cpl_image_basic_body.h"
#undef CPL_CLASS
#define CPL_CLASS CPL_CLASS_FLOAT
#include "cpl_image_basic_body.h"
#undef CPL_CLASS
#define CPL_CLASS CPL_CLASS_INT
#include "cpl_image_basic_body.h"
#undef CPL_CLASS
        default:
            cpl_ensure_code(0, CPL_ERROR_TYPE_MISMATCH) ;
    }

    /* Handle the bad pixels */
    if (im->bpm != NULL) cpl_mask_move(im->bpm, nb_cut, new_pos) ;
    
    return CPL_ERROR_NONE ;

}
#undef CPL_OPERATION

/*----------------------------------------------------------------------------*/
/**
  @brief    Apply a gaussian fit on an image sub window
  @param    im      the input image
  @param    xpos    the x position of the center (1 for the first pixel)
  @param    ypos    the y position of the center (1 for the first pixel)
  @param    size    the window size in pixels
  @param    norm    the norm of the gaussian or NULL
  @param    xcen    the x center of the gaussian or NULL
  @param    ycen    the y center of the gaussian or NULL
  @param    sig_x   the sigma in x of the gaussian or NULL
  @param    sig_y   the sigma in y of the gaussian or NULL
  @param    fwhm_x  the FHHM in x or NULL
  @param    fwhm_y  the FHHM in y or NULL
  @return   the #_cpl_error_code_ or CPL_ERROR_NONE 

  The computed norm, xcen, ycen, sig_x, sig_y coefficients are defining the 
  gaussian:
 
  f(x, y) = (norm/(2*pi*sig_x*sig_y)) * exp(-(x-xcen)^2/(2*sig_x^2)) * 
            exp(-(y-ycen)^2/(2*sig_y^2))

  fwhm_x and fwhm_y are derived from sig_x and sig_y like:
  fwhm = 2 * sqrt(2*ln(2)) * sigma

  Images can be CPL_TYPE_INT, CPL_TYPE_FLOAT or CPL_TYPE_DOUBLE.

  Possible #_cpl_error_code_ set in this function:
  - CPL_ERROR_NULL_INPUT if the input image pointer is NULL
  - CPL_ERROR_ILLEGAL_INPUT if xpos, ypos or size or illegal
  - CPL_ERROR_TYPE_MISMATCH if the passed image type is not supported
*/
/*----------------------------------------------------------------------------*/
cpl_error_code cpl_image_fit_gaussian(
        const cpl_image *   im,
        int                 xpos,
        int                 ypos,
        int                 size,
        double          *   norm,
        double          *   xcen,
        double          *   ycen,
        double          *   sig_x,
        double          *   sig_y,
        double          *   fwhm_x,
        double          *   fwhm_y) 
{
    cpl_image   *   extracted ;
    int             is_rejected;
    int             llx, lly, urx, ury ;
    double          u0, ux, uy, uxx, uyy ;
    double          cenx, ceny;
    const double *  pi ;
    int             pos ;
    double          max_val ;
    int             i, j ;

    /* Check entries */
    cpl_ensure_code(im, CPL_ERROR_NULL_INPUT) ;
    cpl_ensure_code(xpos>=1 && xpos<=im->nx, CPL_ERROR_ILLEGAL_INPUT);
    cpl_ensure_code(ypos>=1 && ypos<=im->ny, CPL_ERROR_ILLEGAL_INPUT);
    cpl_ensure_code(size>1 && size<im->nx && size<im->ny,
            CPL_ERROR_ILLEGAL_INPUT) ;

    /* Extraction zone */
    llx = xpos - (int)(size/2) ;
    lly = ypos - (int)(size/2) ;
    urx = xpos + (int)(size/2) ;
    ury = ypos + (int)(size/2) ;
    if (llx < 1) llx = 1 ;
    if (lly < 1) lly = 1 ;
    if (urx > im->nx) urx = im->nx ;
    if (ury > im->ny) ury = im->ny ;
    
    /* Extract the image zone to fit */
    extracted = cpl_image_extract(im, llx, lly, urx, ury) ;
    cpl_ensure_code(extracted, CPL_ERROR_ILLEGAL_INPUT) ;
  
    /* Check if there are enough good pixels */
    if (5 * cpl_image_count_rejected(extracted) > 
            cpl_image_get_size_x(extracted) * cpl_image_get_size_y(extracted)) {
        cpl_image_delete(extracted) ;
        cpl_ensure_code(0, CPL_ERROR_ILLEGAL_INPUT) ;
    }
    
    if (cpl_image_get_type(extracted) != CPL_TYPE_DOUBLE) {
        /* Convert the image to double */
        cpl_image * tmp = extracted;
        extracted = cpl_image_cast(tmp, CPL_TYPE_DOUBLE);
        cpl_image_delete(tmp);
        cpl_ensure_code(extracted, CPL_ERROR_TYPE_MISMATCH);
    }
    pi = (double*)extracted->pixels ;

    /* Compute xcen and ycen */
    u0 = ux = uy = 0.0 ;
    for (j=0 ; j<extracted->ny ; j++) {
        for (i=0 ; i<extracted->nx ; i++) {
            if (!cpl_image_is_rejected(extracted, i+1, j+1)) {
                pos = i + j * extracted->nx ;
                u0 += pi[pos] ;
                ux += (i+1) * pi[pos] ;
                uy += (j+1) * pi[pos] ;
            }
        }
    }

    /* cenx = ux/u0 may not be outside 1 and nx and
       ceny = uy/u0 may not be outside 1 and ny */
    if (u0 == 0 || u0 > ux || ux > u0*extracted->nx ||
        u0 > uy || uy > u0*extracted->ny) {
        cpl_image_delete(extracted) ;
        cpl_ensure_code(0, CPL_ERROR_ILLEGAL_INPUT) ;
    }

    cenx = ux/u0;
    ceny = uy/u0;
    
    /* Compute sig_x and sig_y */
    uxx = uyy = 0.0 ;
    for (j=0 ; j<extracted->ny ; j++) {
        for (i=0 ; i<extracted->nx ; i++) {
            if (!cpl_image_is_rejected(extracted, i+1, j+1)) {
                pos = i + j * extracted->nx ;
                uxx += ((i+1)-cenx) * ((i+1)-cenx) * pi[pos] ;
                uyy += ((j+1)-ceny) * ((j+1)-ceny) * pi[pos] ;
            }
        }
    }
    if (sig_x) *sig_x = sqrt(fabs(uxx/u0)) ;
    if (sig_y) *sig_y = sqrt(fabs(uyy/u0)) ;
    if (fwhm_x) *fwhm_x = 2 * sqrt(2 * log(2.0)) * sqrt(fabs(uxx/u0)) ;
    if (fwhm_y) *fwhm_y = 2 * sqrt(2 * log(2.0)) * sqrt(fabs(uyy/u0)) ;

    /* Compute norm */
    max_val = cpl_image_get(extracted, (int)cenx, (int)ceny,
                            &is_rejected);
    cpl_ensure_code(is_rejected >= 0, cpl_error_get_code());

    if (is_rejected) {
        cpl_error_push();
        max_val = cpl_image_get_mean_window(extracted, (int)cenx,
                (int)ceny,  (int)(cenx+1), (int)(ceny+1)) ;
        cpl_ensure_code(!cpl_error_get_code(), cpl_error_get_code());
        cpl_error_pop();
    }

    cpl_image_delete(extracted) ;

    if (norm) *norm = max_val*2*CX_PI*sqrt(fabs(uxx/u0))*sqrt(fabs(uyy/u0)) ; 
    
    /* Shift xcen and ycen to coordinates in the input big image */
    if (xcen) *xcen = cenx + llx - 1 ;
    if (ycen) *ycen = ceny + lly - 1 ;
    

    return CPL_ERROR_NONE ;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Compute FWHM values in x and y for an object
  @param    in      the input image
  @param    xpos    the x position of the object (1 for the first pixel)
  @param    ypos    the y position of the object (1 for the first pixel)
  @param    fwhm_x  the computed FWHM in x or -1 on error
  @param    fwhm_y  the computed FWHM in y or -1 on error
  @return   CPL_ERROR_NONE or the relevant #_cpl_error_code_

  This function uses a basic method: start from the center of the object
  and go away until the half maximum value is reached in x and y.

  For the FWHM in x (resp. y) to be computed, the image size in the x (resp. y)
  direction should be at least of 5 pixels.
  
  If for any reason, one of the FHWMs cannot be computed, its returned value 
  is -1.0, but an error is not necessarily raised. For example, if a 4 column
  image is passed, the fwhm_x would be -1.0, the fwhm_y would be correctly 
  computed, and no error would be raised.
  
  Possible #_cpl_error_code_ set in this function:
  - CPL_ERROR_NULL_INPUT if (one of) the input pointer(s) is NULL
  - CPL_ERROR_DATA_NOT_FOUND if (xpos, ypos) specifies a rejected pixel or a
      pixel with a non-positive value
  - CPL_ERROR_ACCESS_OUT_OF_RANGE if xpos or ypos is outside the image
    size range
 */
/*----------------------------------------------------------------------------*/
cpl_error_code cpl_image_get_fwhm(
        const cpl_image *   in,
        int                 xpos,
        int                 ypos,
        double          *   fwhm_x,
        double          *   fwhm_y)
{
    double              half_max;
    double              thres;
    const int           minimum_size = 5;
    int                 is_rejected;
    /* FIXME: cpl_error_push/pop() has only one level, so try to use it only if
       it has not already been used.
       NB: This approach can only be used when there is no recovery from errors
           occuring during the push()/pop(). */
    const cpl_boolean   do_push = cpl_error_get_code() ? CPL_TRUE : CPL_FALSE;


    /* Check entries - and initialize *fwhm_{x,y} */
    if (fwhm_y != NULL) *fwhm_y = -1;
    cpl_ensure_code(fwhm_x,         CPL_ERROR_NULL_INPUT);
    *fwhm_x = -1;
    cpl_ensure_code(fwhm_y,         CPL_ERROR_NULL_INPUT);

    /* This call will check the validity of image, xpos and ypos */
    half_max = 0.5 * cpl_image_get(in, xpos, ypos, &is_rejected);

    cpl_ensure_code(is_rejected >= 0, cpl_error_get_code());
    cpl_ensure_code(!is_rejected, CPL_ERROR_DATA_NOT_FOUND);

    cpl_ensure_code(half_max > 0, CPL_ERROR_DATA_NOT_FOUND);
    
    /* FWHM in x */
    if (in->nx >= minimum_size) {

        /* Extract the vector centered on the maximum */
        cpl_vector * row = cpl_vector_new_from_image_row(in, ypos);

        /* If an error happened, update its location */
        cpl_ensure_code(row, cpl_error_get_code());

        if (do_push) cpl_error_push();

        /* Find out threshold */
        thres = cpl_vector_get_noise(row, xpos);
        
        /* Compute the FWHM */
        if (!cpl_error_get_code())
            *fwhm_x = cpl_vector_get_fwhm(row, xpos, half_max + thres * 0.5);

        cpl_vector_delete(row);

        /* If an error happened, update its location */
        cpl_ensure_code(!cpl_error_get_code(), cpl_error_get_code());

        if (do_push) cpl_error_pop();

    }
     
    /* FWHM in y */
    if (in->ny >= minimum_size) {
        /* Extract the vector centered on the maximum */
        cpl_vector * col = cpl_vector_new_from_image_column(in, xpos);

        /* If an error happened, update its location */
        cpl_ensure_code(col, cpl_error_get_code());

        if (do_push) cpl_error_push();

        /* Find out threshold */
        thres = cpl_vector_get_noise(col, ypos);
        
        /* Compute the FWHM */
        if (!cpl_error_get_code())
            *fwhm_y = cpl_vector_get_fwhm(col, ypos, half_max + thres * 0.5);

        cpl_vector_delete(col);

        /* If an error happened, update its location */
        cpl_ensure_code(!cpl_error_get_code(), cpl_error_get_code());

        if (do_push) cpl_error_pop();

    }
        
    return CPL_ERROR_NONE;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Fast Fourier Transform an image.
  @param    real        The image real part to be transformed in place
  @param    imaginary   The image imaginary part to be transformed in place
  @param    mode   The desired FFT options (combined with bitwise or)
  @return   CPL_ERROR_NONE or the relevant #_cpl_error_code_ on error

  The input images must be of double type.
 
  If the second passed image is NULL, the resulting imaginary part
  cannot be returned. This can be useful if the input is real and the
  output is known to also be real. But if the output has a significant
  imaginary part, you might want to pass a 0-valued image as the second
  parameter.
  
  Any rejected pixel us used as if it were a good pixel.

  The image must be square with a size that is a power of two.

  These are the supported FFT modes:
  CPL_FFT_DEFAULT: Default, forward FFT transform
  CPL_FFT_INVERSE: Inverse FFT transform
  CPL_FFT_UNNORMALIZED: Do not normalize (with N*N for N-by-N image) on inverse.
    Has no effect on forward transform.
  CPL_FFT_SWAP_HALVES: Swap the four quadrants of the result image.

  Possible #_cpl_error_code_ set in this function:
  - CPL_ERROR_NULL_INPUT if (one of) the input pointer(s) is NULL
  - CPL_ERROR_ILLEGAL_INPUT if the image is not square or if the image size is
    not a power of 2.
  - CPL_ERROR_INVALID_TYPE if mode is 1, e.g. due to a logical or (||) of the
      allowed FFT options.
  - CPL_ERROR_UNSUPPORTED_MODE if mode is otherwise different from the allowed
       FFT options.
  - CPL_ERROR_TYPE_MISMATCH if the passed image type is not supported
 */
/*----------------------------------------------------------------------------*/
cpl_error_code cpl_image_fft(
        cpl_image   *   real, 
        cpl_image   *   imaginary, 
        unsigned        mode)
{
    unsigned                dim[2];
    double              *   imag_part ;

    /* Check entries */
    cpl_ensure_code(real, CPL_ERROR_NULL_INPUT);
    cpl_ensure_code(mode != 1, CPL_ERROR_INVALID_TYPE);
    cpl_ensure_code(
            mode <= (CPL_FFT_INVERSE|CPL_FFT_UNNORMALIZED|CPL_FFT_SWAP_HALVES), 
            CPL_ERROR_UNSUPPORTED_MODE);
    cpl_ensure_code(real->nx==real->ny, CPL_ERROR_ILLEGAL_INPUT);
    cpl_ensure_code(real->type==CPL_TYPE_DOUBLE,CPL_ERROR_TYPE_MISMATCH);
    cpl_ensure_code(cpl_tools_is_power_of_2(real->nx)>=0,
                    CPL_ERROR_ILLEGAL_INPUT);
    if (imaginary != NULL) {
        cpl_ensure_code(real->nx==imaginary->nx,CPL_ERROR_ILLEGAL_INPUT);
        cpl_ensure_code(real->ny==imaginary->ny,CPL_ERROR_ILLEGAL_INPUT);
        cpl_ensure_code(imaginary->type==CPL_TYPE_DOUBLE,
                CPL_ERROR_TYPE_MISMATCH);
    }

    /* Initialize */
    dim[0] = dim[1] = real->nx;

    if (imaginary == NULL) {
        /* Create the imaginary part and set it to 0  */
        imag_part = cpl_calloc(dim[0]*dim[1], sizeof(double)) ;
    } else {
        /* Put the input imaginary part in a local object */
        imag_part = imaginary->pixels ;
    }

    /* APPLY THE FFT HERE */
    cpl_ensure_code(!cpl_fft(real->pixels, imag_part, dim, 2,
                (mode & CPL_FFT_INVERSE) ? -1 : 1),cpl_error_get_code());

    /* Free the imaginary part result in the input image */
    if (imaginary == NULL) {
        cpl_free(imag_part) ;
    }

    /* Normalize on the inverse transform  */
    if (!(mode & CPL_FFT_UNNORMALIZED) && (mode & CPL_FFT_INVERSE)) {
        cpl_ensure_code(!cpl_image_divide_scalar(real, dim[0]*dim[1]),
                        cpl_error_get_code());
        if (imaginary != NULL) {
            cpl_ensure_code(!cpl_image_divide_scalar(imaginary, dim[0]*dim[1]),
                            cpl_error_get_code());
        }
    }

    /* Swap halves in both dimensions */
    if (mode & CPL_FFT_SWAP_HALVES) {
        const int new_pos[] = {4, 3, 2, 1};
        cpl_ensure_code(!cpl_image_move(real, 2, new_pos),
                        cpl_error_get_code());
        if (imaginary != NULL) {
            cpl_ensure_code(!cpl_image_move(imaginary, 2, new_pos),
                            cpl_error_get_code());
        }
    }

    return CPL_ERROR_NONE;
}
/**@}*/

/*----------------------------------------------------------------------------*/
/**
  @brief    N-dimensional FFT.
  @param    real        N-dimensional data set stored in 1d (real part).
  @param    imag        N-dimensional data set stored in 1d (imaginary part).
  @param    nn          Dimensions of the set.
  @param    ndim        How many dimensions this set has.
  @param    isign       Transform direction.
  @return   the #_cpl_error_code_ or CPL_ERROR_NONE

  This routine is a public domain FFT. See extract of Usenet article below. 
  Found on http://www.tu-chemnitz.de/~arndt/joerg.html

  @verbatim
  From: alee@tybalt.caltech.edu (Andrew Lee)
  Newsgroups: comp.sources.misc
  Subject: N-dimensional, Radix 2 FFT Routine
  Date: 17 Jul 87 22:26:29 GMT
  Approved: allbery@ncoast.UUCP
  X-Archive: comp.sources.misc/8707/48

  [..]
  Now for the usage (finally):
  real[] and imag[] are the array of complex numbers to be transformed,
  nn[] is the array giving the dimensions (I mean size) of the array,
  ndim is the number of dimensions of the array, and
  isign is +1 for a forward transform, and -1 for an inverse transform.

  real[], imag[] and nn[] are stored in the "natural" order for C:
  nn[0] gives the number of elements along the leftmost index,
  nn[ndim - 1] gives the number of elements along the rightmost index.

  Additional notes: The routine does NO NORMALIZATION, so if you do a forward, 
  and then an inverse transform on an array, the result will be identical to 
  the original array MULTIPLIED BY THE NUMBER OF ELEMENTS IN THE ARRAY. Also, 
  of course, the dimensions of real[] and imag[] must all be powers of 2.
  @endverbatim

  Possible #_cpl_error_code_ set in this function:
  - CPL_ERROR_NULL_INPUT if (one of) the input pointer(s) is NULL
 */
/*----------------------------------------------------------------------------*/
static cpl_error_code cpl_fft(
        double          *   real,
        double          *   imag,
        const unsigned  *   nn,
        int                 ndim,
        int                 isign)
{
    int                 idim;
    unsigned            i1, i2rev, i3rev, ibit;
    unsigned            ip2, ifp1, ifp2, k2, n;
    unsigned            nprev = 1, ntot = 1;
    register unsigned   i2, i3;
    double              theta;
    double              w_r, w_i, wp_r, wp_i;
    double              wtemp;
    double              temp_r, temp_i, wt_r, wt_i;
    double              t1, t2;

    /* Check entries */
    cpl_ensure_code(real, CPL_ERROR_NULL_INPUT);
    cpl_ensure_code(imag, CPL_ERROR_NULL_INPUT);
    cpl_ensure_code(nn, CPL_ERROR_NULL_INPUT);
    for (idim = 0; idim < ndim; ++idim) 
        cpl_ensure_code(cpl_tools_is_power_of_2(nn[idim])>=0,
                        CPL_ERROR_ILLEGAL_INPUT);

    /* Compute total number of complex values  */
    for (idim = 0; idim < ndim; ++idim) ntot *= nn[idim];

    for (idim = ndim - 1; idim >= 0; --idim) {
        n = nn[idim];

        ip2 = nprev * n;        /*  Unit step for next dimension */
        i2rev = 0;              /*  Bit reversed i2 */

        /* This is the bit reversal section of the routine */
        /* Loop over current dimension */
        for (i2 = 0; i2 < ip2; i2 += nprev) {
            if (i2 < i2rev)
                /*      Loop over lower dimensions      */
                for (i1 = i2; i1 < i2 + nprev; ++i1)
                    /*      Loop over higher dimensions  */
                    for (i3 = i1; i3 < ntot; i3 += ip2) {
                        i3rev = i3 + i2rev - i2;
                        temp_r = real[i3];
                        temp_i = imag[i3];
                        real[i3] = real[i3rev];
                        imag[i3] = imag[i3rev];
                        real[i3rev] = temp_r;
                        imag[i3rev] = temp_i;
                    }

            ibit = ip2;
            /* Increment from high end of i2rev to low */
            do {
                ibit >>= 1;
                i2rev ^= ibit;
            } while (ibit >= nprev && !(ibit & i2rev));
        }

        /* Here begins the Danielson-Lanczos section of the routine */
        /* Loop over step sizes    */
        for (ifp1 = nprev; ifp1 < ip2; ifp1 <<= 1) {
            ifp2 = ifp1 << 1;
            /*  Initialize for the trig. recurrence */
            theta = isign * 2.0 * CX_PI / (ifp2 / nprev);
            wp_r = sin(0.5 * theta);
            wp_r *= -2.0 * wp_r;
            wp_i = sin(theta);
            w_r = 1.0;
            w_i = 0.0;

            /* Loop by unit step in current dimension  */
            for (i3 = 0; i3 < ifp1; i3 += nprev) {
                /* Loop over lower dimensions      */
                for (i1 = i3; i1 < i3 + nprev; ++i1)
                    /* Loop over higher dimensions */
                    for (i2 = i1; i2 < ntot; i2 += ifp2) {
                        /* Danielson-Lanczos formula */
                        k2 = i2 + ifp1;
                        wt_r = real[k2];
                        wt_i = imag[k2];

                        /* Complex multiply using 3 real multiplies. */
                        real[k2] = real[i2] - (temp_r =
                            (t1 = w_r * wt_r) - (t2 = w_i * wt_i));
                        imag[k2] = imag[i2] - (temp_i =
                            (w_r + w_i) * (wt_r + wt_i) - t1 - t2);
                        real[i2] += temp_r;
                        imag[i2] += temp_i;
                    }
                /* Trigonometric recurrence */
                wtemp = w_r;
                /* Complex multiply using 3 real multiplies. */
                w_r += (t1 = w_r * wp_r) - (t2 = w_i * wp_i);
                w_i += (wtemp + w_i) * (wp_r + wp_i) - t1 - t2;
            }
        }
        nprev *= n;
    }
    return CPL_ERROR_NONE;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Compute the noise around a peak in a vector
  @param    vec     the input cpl_vector
  @param    pos     the peak position (from 1 to the vector size)
  @return   the noise value.

  The passed cpl_vector object must have at least two elements.
  
  In case of error, the #_cpl_error_code_ code is set, and the returned double
  is undefined.

  Possible #_cpl_error_code_ set in this function:
  - CPL_ERROR_NULL_INPUT if (one of) the input pointer(s) is NULL
  - CPL_ERROR_ILLEGAL_INPUT
 */
/*----------------------------------------------------------------------------*/
static double cpl_vector_get_noise(
        const cpl_vector    *   vec,
        int                     pos)
{
    int                 nelem ;
    cpl_vector       *  smooth_vec ;
    const double     *  vector_data ;
    double              noise_left, noise_right ;
    int                 i ;

    /* Check entries */
    cpl_ensure(vec,          CPL_ERROR_NULL_INPUT, -1.0) ;
    nelem = cpl_vector_get_size(vec) ;
    cpl_ensure(pos >= 1,     CPL_ERROR_ILLEGAL_INPUT, -1.0) ;
    cpl_ensure(pos <= nelem, CPL_ERROR_ILLEGAL_INPUT, -1.0) ;
    cpl_ensure(nelem > 1,    CPL_ERROR_ILLEGAL_INPUT, -1.0) ;

    /* Smooth out the array to be less sensitive to noise */
    smooth_vec = cpl_vector_filter_lowpass_create(vec, CPL_LOWPASS_LINEAR, 1) ;
    /* If that fails, use the unsmoothed vector */
    vector_data = cpl_vector_get_data(smooth_vec ? smooth_vec : vec);

    /* Find noise level on the left side of the peak. */
    i = pos - 1 ;
    while (i > 0) {
        if (vector_data[i] > vector_data[i-1]) i--;
        else break;
    }
    noise_left = vector_data[i] ;

    /* Find noise level on the right side of the peak */
    i = pos - 1 ;
    while (i < nelem-1) {
        if (vector_data[i] > vector_data[i+1]) i++ ;
        else break ;
    }
    noise_right = vector_data[i] ;

    cpl_vector_delete(smooth_vec) ;

    return 0.5 * (noise_left + noise_right);
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Compute the FWHM of an object in a cpl_vector
  @param    vec         the input cpl_vector
  @param    pos         the object position (from 1 to the vector size)
  @param    half_max    the half maximum value
  @return   the FWHM or a negative value in error case
  @note The return value may be -1 with no error condition

  Possible #_cpl_error_code_ set in this function:
  - CPL_ERROR_NULL_INPUT if an input pointer is NULL
  - CPL_ERROR_ILLEGAL_INPUT if pos is less than 1 or greater than the vec-length
 */
/*----------------------------------------------------------------------------*/
static double cpl_vector_get_fwhm(
        const cpl_vector    *   vec,
        int                     pos,
        double                  half_max)
{
    const double * vec_data;
    int            nelem;
    double         x_left, x_right;
    double         y1, y2;
    int            i;

    /* Check entries */
    cpl_ensure(vec,          CPL_ERROR_NULL_INPUT, -1.0) ;
    nelem = cpl_vector_get_size(vec) ;
    cpl_ensure(pos >= 1,     CPL_ERROR_ILLEGAL_INPUT, -1.0) ;
    cpl_ensure(pos <= nelem, CPL_ERROR_ILLEGAL_INPUT, -1.0) ;

    vec_data = cpl_vector_get_data(vec) ;

    /* Object may be too noisy - or strange in some other way */
    if (vec_data[pos - 1] <= half_max) return -1.0;

    /* Find first pair of values, y(i) <= half_max < y(i+1)
         on the left of the maximum */
    i = pos - 1;

    while ((vec_data[i] > half_max) && (i > 0)) i--;
    if (vec_data[i] > half_max) return -1.0;  /* y1 could not be found */

    y1 = vec_data[i] ;
    y2 = vec_data[i+1] ;

    /* assert ( y1 <= half_max && half_max < y2 ); */

    /* Assume linearity between y1 and y2 */
    x_left = i + (half_max-y1) / (y2-y1) ;

    /* assert( x_left >= i ); */

    /* Find first pair of values, y(i-1) > half_max >= y(i)
         on the right of the maximum */
    i = pos - 1 ;
 
    while ((vec_data[i] > half_max) && (i < nelem-1)) i++;
    if (vec_data[i] > half_max) return -1.0;   /* y2 could not be found */

    y1 = vec_data[i-1] ;
    y2 = vec_data[i] ;

    /* assert( y1 > half_max && half_max >= y2 ); */

    /* Assume linearity between y1 and y2 */
    x_right = i + (half_max-y2) / (y2-y1) ;

    /* assert( x_right < i ); */

    if (x_right < x_left || x_right - x_left > FLT_MAX) return -1;

    return x_right - x_left ;
}
