/* $Id: cpl_imagelist_io.c,v 1.39 2006/12/01 13:15:26 yjung Exp $
 *
 * This file is part of the ESO Common Pipeline Library
 * Copyright (C) 2001-2004 European Southern Observatory
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/*
 * $Author: yjung $
 * $Date: 2006/12/01 13:15:26 $
 * $Revision: 1.39 $
 * $Name: cpl-3_1 $
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

/*-----------------------------------------------------------------------------
                                Includes
 -----------------------------------------------------------------------------*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <math.h>
#include <assert.h>

#include <qfits.h>

#include "cpl_tools.h"
#include "cpl_propertylist_impl.h"
#include "cpl_memory.h"
#include "cpl_imagelist_io.h"

#include "cpl_imagelist_defs.h"

/*-----------------------------------------------------------------------------
                                Defines
 -----------------------------------------------------------------------------*/

#define CPL_IMLIST_IO_LOAD      0
#define CPL_IMLIST_IO_LOAD_WIN  1
#define CPL_IMLIST_IO_SAVE      2

/*----------------------------------------------------------------------------*/
/**
 * @defgroup cpl_imagelist_io Image list IO routines
 *
 * This module provides IO functions to handle @em cpl_imagelist.
 *
 * @par Synopsis:
 * @code
 *   #include "cpl_imagelist_io.h"
 * @endcode
 */
/*----------------------------------------------------------------------------*/
/**@{*/

/*-----------------------------------------------------------------------------
                            Function codes
 -----------------------------------------------------------------------------*/

/*----------------------------------------------------------------------------*/
/**
  @brief    Create an empty imagelist
  @return   1 newly allocated cpl_imagelist
  @see      cpl_imagelist_set()
  
  The returned cpl_imagelist must be deallocated using cpl_imagelist_delete()

 */
/*----------------------------------------------------------------------------*/
cpl_imagelist * cpl_imagelist_new()
{
    return (cpl_imagelist *) cpl_calloc(1, sizeof(cpl_imagelist));
}

#define CPL_OPERATION CPL_IMLIST_IO_LOAD
/*----------------------------------------------------------------------------*/
/**
  @brief    Load a FITS file extension into a list of images
  @param    filename    The FITS file name
  @param    type        Type of the images in the created image list
  @param    extnum      The extension number
  @return   The loaded list of images or NULL on error.
  @see      cpl_image_load()
 
  This function loads all the images of a specified extension (NAXIS=2
  or 3) in an image list.
  
  Type can be CPL_TYPE_DOUBLE, CPL_TYPE_FLOAT or CPL_TYPE_INT.
  The loaded images have an empty bad pixel map.

  The returned cpl_imagelist must be deallocated using cpl_imagelist_delete()

  Possible #_cpl_error_code_ set in this function:
  - CPL_ERROR_NULL_INPUT if (one of) the input pointer(s) is NULL
  - CPL_ERROR_ILLEGAL_INPUT if extnum is negative
  - CPL_ERROR_TYPE_MISMATCH if the passed type is not supported
  - CPL_ERROR_ACCESS_OUT_OF_RANGE if extnum is bigger than the number of
    extensions in the FITS file
  - CPL_ERROR_BAD_FILE_FORMAT if the file cannot be parsed
  - CPL_ERROR_DATA_NOT_FOUND if the datacannot be read from the file
 */
/*----------------------------------------------------------------------------*/
cpl_imagelist * cpl_imagelist_load(
        const char * filename,
        cpl_type     type,
        int          extnum)
{

    switch (type) {
    case CPL_TYPE_DOUBLE: {
#define CPL_CLASS CPL_CLASS_DOUBLE
#include "cpl_imagelist_io_body.h"
#undef CPL_CLASS
    }

    case CPL_TYPE_FLOAT: {
#define CPL_CLASS CPL_CLASS_FLOAT
#include "cpl_imagelist_io_body.h"
#undef CPL_CLASS
    }
    case CPL_TYPE_INT: {
#define CPL_CLASS CPL_CLASS_INT
#include "cpl_imagelist_io_body.h"
#undef CPL_CLASS
    }
    default:
        cpl_ensure(0, CPL_ERROR_TYPE_MISMATCH, NULL);
    }
}
#undef CPL_OPERATION


#define CPL_OPERATION CPL_IMLIST_IO_LOAD_WIN
/*----------------------------------------------------------------------------*/
/**
  @brief    Load images windows from a FITS file extension into an image list
  @param    filename    The FITS file name
  @param    type        Type of the images in the created image list
  @param    extnum      The extension number
  @param    llx         Specifies the window position
  @param    lly         Specifies the window position
  @param    urx         Specifies the window position
  @param    ury         Specifies the window position
  @return   The loaded list of image windows or NULL on error.
  @see      cpl_image_load_window()
 
  This function loads all the image windows of a specified extension in an
  image list.
  
  Type can be CPL_TYPE_DOUBLE, CPL_TYPE_FLOAT or CPL_TYPE_INT.

  The window position is specified starting with 1 to the number of
  pixel. Lower left pixel is (1, 1).

  The returned cpl_imagelist must be deallocated using cpl_imagelist_delete()

  Possible #_cpl_error_code_ set in this function:
  - CPL_ERROR_NULL_INPUT if (one of) the input pointer(s) is NULL
  - CPL_ERROR_ILLEGAL_INPUT if extnum is negative
  - CPL_ERROR_TYPE_MISMATCH if the passed type is not supported
  - CPL_ERROR_ACCESS_OUT_OF_RANGE if extnum is bigger than the number of
    extensions in the FITS file
  - CPL_ERROR_BAD_FILE_FORMAT if the file cannot be parsed
  - CPL_ERROR_DATA_NOT_FOUND if the datacannot be read from the file
 */
/*----------------------------------------------------------------------------*/
cpl_imagelist * cpl_imagelist_load_window(
        const char  *   filename,
        cpl_type        type,
        int             extnum,
        int             llx,
        int             lly,
        int             urx,
        int             ury)
{

    switch (type) {
        case CPL_TYPE_DOUBLE: 
            {
#define CPL_CLASS CPL_CLASS_DOUBLE
#include "cpl_imagelist_io_body.h"
#undef CPL_CLASS
            }
        case CPL_TYPE_FLOAT: 
            {
#define CPL_CLASS CPL_CLASS_FLOAT
#include "cpl_imagelist_io_body.h"
#undef CPL_CLASS
            }
        case CPL_TYPE_INT: 
            {
#define CPL_CLASS CPL_CLASS_INT
#include "cpl_imagelist_io_body.h"
#undef CPL_CLASS
            }
        default:
            cpl_ensure(0, CPL_ERROR_TYPE_MISMATCH, NULL);
    }
}
#undef CPL_OPERATION

/*----------------------------------------------------------------------------*/
/**
  @brief    Get the number of images in the imagelist
  @param    imlist    the list of image
  @return   The number of images or -1 on error

  Possible #_cpl_error_code_ set in this function:
  - CPL_ERROR_NULL_INPUT if (one of) the input pointer(s) is NULL
 */
/*----------------------------------------------------------------------------*/
int cpl_imagelist_get_size(const cpl_imagelist * imlist)
{
    cpl_ensure(imlist != NULL, CPL_ERROR_NULL_INPUT,
               -1);

    assert( imlist->ni >= 0 );

    return imlist->ni ;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Get an image from a list of images
  @param    imlist    the image list
  @param    inum    the image id (from 0 to number of images-1)
  @return   A pointer to the image or NULL in error case.

  The returned pointer refers to already allocated data.

  Possible #_cpl_error_code_ set in this function:
  - CPL_ERROR_NULL_INPUT if (one of) the input pointer(s) is NULL
  - CPL_ERROR_ACCESS_OUT_OF_RANGE if inum is bigger thant the list size
  - CPL_ERROR_ILLEGAL_INPUT if inum is negative
 */
/*----------------------------------------------------------------------------*/
cpl_image * cpl_imagelist_get(
        const cpl_imagelist * imlist,
        int                   inum)
{

    cpl_ensure(imlist != NULL,    CPL_ERROR_NULL_INPUT, NULL);
    cpl_ensure(inum >= 0,         CPL_ERROR_ILLEGAL_INPUT, NULL);
    cpl_ensure(inum < imlist->ni, CPL_ERROR_ACCESS_OUT_OF_RANGE, NULL);
    
    return imlist->images[inum] ;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Insert an image into an imagelist
  @param    imlist   The imagelist
  @param    im       The image to insert
  @param    pos      The list position (from 0 to number of images)
  @return   CPL_ERROR_NONE or the relevant #_cpl_error_code_ on error

  The image is put at the position pos in the image list. The image that is
  already there is deallocated.

  It is allowed to specify the position equal to the number of images in the
  list. This will increment the size of the imagelist.

  No action occurs if an image is inserted more than once into the same
  position. It is not allowed to insert the same image into two different 
  positions in a list.

  It is not allowed to insert images of different size into a list.

  You should not deallocate the image after it has been inserted into an
  imagelist, since the cpl_imagelist_delete() will deallocate its images.
  
  Possible #_cpl_error_code_ set in this function:
  - CPL_ERROR_NULL_INPUT if (one of) the input pointer(s) is NULL
  - CPL_ERROR_ILLEGAL_INPUT if pos is negative
  - CPL_ERROR_TYPE_MISMATCH if im and imlist are of different types
  - CPL_ERROR_INCOMPATIBLE_INPUT if im and imlist have different sizes
  - CPL_ERROR_ACCESS_OUT_OF_RANGE if pos is bigger than the number of
    images in imlist
 */
/*----------------------------------------------------------------------------*/
cpl_error_code cpl_imagelist_set(
        cpl_imagelist * imlist, 
        cpl_image     * im, 
        int             pos)
{
    int                 i;

    cpl_ensure_code(imlist,            CPL_ERROR_NULL_INPUT);
    cpl_ensure_code(im,                CPL_ERROR_NULL_INPUT);
    cpl_ensure_code(pos >= 0,          CPL_ERROR_ILLEGAL_INPUT);
    cpl_ensure_code(pos <= imlist->ni, CPL_ERROR_ACCESS_OUT_OF_RANGE);

    /* Do nothing if the image is already there */
    if (pos < imlist->ni && im == imlist->images[pos]) return CPL_ERROR_NONE;

    /* Do not allow the image to be inserted into two different positions */
    for (i = 0; i < imlist->ni; i++) 
        cpl_ensure_code(im != imlist->images[i], CPL_ERROR_ILLEGAL_INPUT);

    if (pos > 0 || imlist->ni > 0) {
        /* Require images to have the same size and type */
        cpl_ensure_code(cpl_image_get_size_x(im) ==
                        cpl_image_get_size_x(imlist->images[0]),
                        CPL_ERROR_INCOMPATIBLE_INPUT);
        cpl_ensure_code(cpl_image_get_size_y(im) ==
                        cpl_image_get_size_y(imlist->images[0]),
                        CPL_ERROR_INCOMPATIBLE_INPUT);
        cpl_ensure_code(cpl_image_get_type(im) ==
                        cpl_image_get_type(imlist->images[0]),
                        CPL_ERROR_TYPE_MISMATCH);
    }

    if (pos == imlist->ni) {
        imlist->ni++;
        imlist->images = cpl_realloc(imlist->images,
                                 imlist->ni * sizeof(cpl_image*));
    } else {
        /* Delete any existing image at that position */
        cpl_image_delete(imlist->images[pos]);
    }

    imlist->images[pos] = im ;

    return CPL_ERROR_NONE ;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Remove an image from an imagelist
  @param    imlist   The imagelist
  @param    pos      The list position (from 0 to number of images-1)
  @return   The pointer to the removed image or NULL in error case

  The specified image is not deallocated, it is simply removed from the
  list. The pointer to the image is returned to let the user decide to
  deallocate it or not.
  Eventually, the image will have to be deallocated with cpl_image_delete().
  
  Possible #_cpl_error_code_ set in this function:
  - CPL_ERROR_NULL_INPUT if (one of) the input pointer(s) is NULL
  - CPL_ERROR_ILLEGAL_INPUT if pos is negative
  - CPL_ERROR_ACCESS_OUT_OF_RANGE if pos is bigger than the number of
    images in imlist
 */
/*----------------------------------------------------------------------------*/
cpl_image * cpl_imagelist_unset(
        cpl_imagelist * imlist, 
        int             pos)
{
    cpl_image       **  new_images ;
    cpl_image       *   out ;
    int                 cur_ind ;
    int                 i ;

    /* Check entries */
    cpl_ensure(imlist, CPL_ERROR_NULL_INPUT, NULL) ;
    cpl_ensure(pos >= 0, CPL_ERROR_ILLEGAL_INPUT, NULL) ;
    cpl_ensure(pos < imlist->ni, CPL_ERROR_ACCESS_OUT_OF_RANGE, NULL) ;

    /* Create the new images container */
    new_images = cpl_malloc((imlist->ni-1) * sizeof(cpl_image*)) ;

    /* Assign the images to the new container */
    cur_ind = 0 ;
    for (i=0 ; i<imlist->ni ; i++) {
        if (i != pos) {
            new_images[cur_ind] = imlist->images[i] ;
            cur_ind++ ;
        }
    }
    out = imlist->images[pos] ;

    /* Reduce the size */
    imlist->ni-- ;
    imlist->images = new_images ;

    return out ;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Free memory associated to a cpl_imagelist object.
  @param    imlist    the image list
 */
/*----------------------------------------------------------------------------*/
void cpl_imagelist_delete(cpl_imagelist * imlist)
{
    int            i ;

    if (imlist == NULL) return ;
    for (i=0 ; i<imlist->ni ; i++) cpl_image_delete(imlist->images[i]) ;
    if (imlist->images) cpl_free(imlist->images);
    cpl_free(imlist) ;
    return ;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Copy an image list
  @param    imlist    Source image list.
  @return   1 newly allocated image list, or NULL on error.

  Copy an image list into a new image list object. 
  The returned image list must be deallocated using cpl_imagelist_delete().

  Possible #_cpl_error_code_ set in this function:
  - CPL_ERROR_NULL_INPUT if (one of) the input pointer(s) is NULL
 */
/*----------------------------------------------------------------------------*/
cpl_imagelist * cpl_imagelist_duplicate(const cpl_imagelist * imlist)
{
    cpl_imagelist *   out ;
    int               i;

    cpl_ensure(imlist != NULL, CPL_ERROR_NULL_INPUT, NULL) ;
    
    /* Create the new imagelist */
    out = cpl_imagelist_new() ;
    
    /* Duplicate the images */
    for (i=0 ; i<imlist->ni ; i++) {
        cpl_imagelist_set(out, cpl_image_duplicate(imlist->images[i]), i) ;
    }

    return out ;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Reject one or more images in a list according to an array of flags.
  @param    imlist  Non-empty imagelist to examine for image rejection.
  @param    valid   Vector of flags (>=-0.5: valid, <-0.5: invalid)
  @return   CPL_ERROR_NONE or the relevant #_cpl_error_code_ on error

  This function takes an imagelist and a vector of flags. The imagelist and
  vector must have equal lengths.

  Images flagged as invalid are removed from the list.

  The removal of image(s) will reduce the length of the list accordingly.

  Possible #_cpl_error_code_ set in this function:
  - CPL_ERROR_NULL_INPUT if (one of) the input pointer(s) is NULL
  - CPL_ERROR_INCOMPATIBLE_INPUT if the vector size and the image list
    size are different
 */
/*----------------------------------------------------------------------------*/
cpl_error_code cpl_imagelist_erase(
        cpl_imagelist    * imlist,
        const cpl_vector * valid) 
{
    int          nkeep = 0;
    int          i;

    /* Check entries */
    cpl_ensure_code(imlist, CPL_ERROR_NULL_INPUT);
    cpl_ensure_code(valid,  CPL_ERROR_NULL_INPUT);
    cpl_ensure_code(cpl_vector_get_size(valid) == imlist->ni,
                    CPL_ERROR_INCOMPATIBLE_INPUT);

    for (i=0 ; i < imlist->ni ; i++) {
        if (cpl_vector_get(valid, i) >= -0.5) {
            /* image is to be kept, place it in the 1st free position */
            imlist->images[nkeep] = imlist->images[i];
            nkeep++;
        } else {
            /* image is to be erased, delete it */
            cpl_image_delete(imlist->images[i]);
        }
    }

    /* Update the size of the altered list */
    imlist->ni = nkeep;

    return CPL_ERROR_NONE ;
}

#define CPL_OPERATION CPL_IMLIST_IO_SAVE
/*----------------------------------------------------------------------------*/
/**
  @brief    Save an imagelist to disk in FITS format.
  @param    imlist      Imagelist to save
  @param    filename    Name of the FITS file to write
  @param    bpp         Bits per pixels
  @param    pl          Property list for the output header or NULL
  @param    mode        The desired output options (combined with bitwise or)
  @return   the #_cpl_error_code_ or CPL_ERROR_NONE
  @see cpl_image_save()

  This function saves an image list to a FITS file, using qfits. If a property 
  list is provided, it is written to the named file before the pixels are 
  written.
  
  The requested pixel depth (bpp) follows the FITS convention. Possible values
  are CPL_BPP_8_UNSIGNED (8), CPL_BPP_16_SIGNED (16), CPL_BPP_16_UNSIGNED (-16),
  CPL_BPP_32_SIGNED (32), CPL_BPP_IEEE_FLOAT (-32), CPL_BPP_IEEE_DOUBLE (-64) 
  or CPL_BPP_DEFAULT.

  In the case of CPL_BPP_16_UNSIGNED, the BITPIX stays 16, the data are
  stored as signed (32768 is subtracted to the images) and BZERO is set to 
  32768 as specified by the FITS standard.

  Supported image lists types are CPL_TYPE_DOUBLE, CPL_TYPE_FLOAT, CPL_TYPE_INT.
        
  Supported output modes are CPL_IO_DEFAULT (create a new file) and
  CPL_IO_EXTEND  (append to an existing file)
 
  If you are in append mode, make sure that the file has writing
  permissions. You may have problems if you create a file in your
  application and append something to it with the umask set to 222. In
  this case, the file created by your application would not be writable,
  and the append would fail.

  Currently, the only supported output mode is CPL_IO_DEFAULT 
  (create a new file).
    
  Possible #_cpl_error_code_ set in this function:
  - CPL_ERROR_NULL_INPUT if (one of) the input pointer(s) is NULL
  - CPL_ERROR_ILLEGAL_INPUT if the bpp value or the mode is not supported
  - CPL_ERROR_FILE_IO if the file cannot be written
  - CPL_ERROR_TYPE_MISMATCH if the passed image list type is not supported
 */
/*----------------------------------------------------------------------------*/
cpl_error_code cpl_imagelist_save(
        const cpl_imagelist     *   imlist,
        const char              *   filename,
        cpl_type_bpp                bpp,
        const cpl_propertylist  *   pl,
        unsigned                    mode)
{
    char                cval[80] ;
    FILE            *   out ;
    qfitsdumper         qd ;
    qfits_header    *   qhdr ;
    qfits_header    *   qhdr_out ;
    cpl_imagelist   *   outilist ;
    cpl_type_bpp        bpp_loc ;
    int                 nx, ny ;
    int                 i ;

    /* Test entries */
    cpl_ensure_code(filename, CPL_ERROR_NULL_INPUT);
    cpl_ensure_code(imlist, CPL_ERROR_NULL_INPUT);
    cpl_ensure_code(bpp==CPL_BPP_8_UNSIGNED  || bpp==CPL_BPP_16_SIGNED ||
                    bpp==CPL_BPP_16_UNSIGNED || bpp==CPL_BPP_32_SIGNED ||
                    bpp==CPL_BPP_IEEE_FLOAT  || bpp==CPL_BPP_IEEE_DOUBLE,
                    CPL_ERROR_ILLEGAL_INPUT);
    cpl_ensure_code(mode <= CPL_IO_MAX, CPL_ERROR_ILLEGAL_INPUT) ;
    /* FIXME: Allow append */
    cpl_ensure_code(mode == CPL_IO_DEFAULT, CPL_ERROR_ILLEGAL_INPUT) ;

    cpl_ensure_code(cpl_imagelist_is_uniform(imlist)==0,
                    CPL_ERROR_ILLEGAL_INPUT);
    nx = cpl_image_get_size_x(cpl_imagelist_get(imlist, 0));
    ny = cpl_image_get_size_y(cpl_imagelist_get(imlist, 0));

    /* Open the file */
    cpl_ensure_code((out = fopen(filename, "w")), CPL_ERROR_FILE_IO);
    
    /* Handle CPL_BPP_16_UNSIGNED */
    if (bpp == CPL_BPP_16_UNSIGNED) {
        bpp_loc = CPL_BPP_16_SIGNED ;
    } else {
        bpp_loc = bpp ;
    }

    /* Create a minimal header */
    qhdr = qfits_header_default();
    sprintf(cval, "%d", bpp_loc);
    qfits_header_add(qhdr, "BITPIX", cval, "bits per pixel", NULL);
    if (imlist->ni > 1) {
        qfits_header_add_after(qhdr, "BITPIX", "NAXIS", "3", "data cube", NULL);
        sprintf(cval, "%d", nx);
        qfits_header_add_after(qhdr, "NAXIS",  "NAXIS1", cval, "x size", NULL);
        sprintf(cval, "%d", ny);
        qfits_header_add_after(qhdr, "NAXIS1", "NAXIS2", cval, "y size", NULL);
        sprintf(cval, "%d", imlist->ni);
        qfits_header_add_after(qhdr, "NAXIS2", "NAXIS3", cval, "z size", NULL);
    } else {
        qfits_header_add_after(qhdr, "BITPIX","NAXIS","2","single image",NULL);
        sprintf(cval, "%d", nx);
        qfits_header_add_after(qhdr, "NAXIS",  "NAXIS1", cval, "x size", NULL);
        sprintf(cval, "%d", ny);
        qfits_header_add_after(qhdr, "NAXIS1", "NAXIS2", cval, "y size", NULL);
    }
    qfits_header_add(qhdr, "BSCALE", "1.0", "pixel scale factor", NULL);

    /* Handle CPL_BPP_16_UNSIGNED */
    if (bpp == CPL_BPP_16_UNSIGNED) {
        qfits_header_add(qhdr, "BZERO", "32768.0", "Offset factor", NULL);
        outilist = cpl_imagelist_duplicate(imlist) ;
        cpl_imagelist_subtract_scalar(outilist, 32768.0) ;
    } else {
        qfits_header_add(qhdr, "BZERO", "0.0", "Offset factor", NULL);
        outilist = (cpl_imagelist *)imlist ;
    }
    /* Append the keywords from pl */
    qhdr_out = cpl_tools_gen_valid_header(qhdr, pl, 0) ;
    qfits_header_destroy(qhdr);

    /* Dump the header */
    qfits_header_sort(&qhdr_out) ;
    qfits_header_dump(qhdr_out, out);
    qfits_header_destroy(qhdr_out);
    fclose(out);

    /* Convert planes one by one and copy them into Fits structure  */
    for (i=0 ; i<outilist->ni ; i++) {
        qd.filename  = (char*) filename ;
        qd.npix      = nx * ny ;
        qd.out_ptype = bpp_loc ;
        switch (cpl_image_get_type(outilist->images[i])) {
#define CPL_CLASS CPL_CLASS_DOUBLE
#include "cpl_imagelist_io_body.h"
#undef CPL_CLASS

#define CPL_CLASS CPL_CLASS_FLOAT
#include "cpl_imagelist_io_body.h"
#undef CPL_CLASS

#define CPL_CLASS CPL_CLASS_INT
#include "cpl_imagelist_io_body.h"
#undef CPL_CLASS
            default:
                if (bpp == CPL_BPP_16_UNSIGNED) cpl_imagelist_delete(outilist) ;
                cpl_ensure(0, CPL_ERROR_TYPE_MISMATCH, CPL_ERROR_TYPE_MISMATCH);
        }
        cpl_ensure(qfits_pixdump(&qd)==0, CPL_ERROR_FILE_IO, CPL_ERROR_FILE_IO);
    }
    if (bpp == CPL_BPP_16_UNSIGNED) cpl_imagelist_delete(outilist) ;
    qfits_zeropad(filename) ;

    return CPL_ERROR_NONE;
}
#undef CPL_OPERATION


/*----------------------------------------------------------------------------*/
/**
  @brief    Determine if an imagelist contains images of equal size and type
  @param    imlist    The imagelist to check
  @return   Zero if uniform, positive if non-uniform and negative on error.

  The function returns 1 if the list is empty.

  Possible #_cpl_error_code_ set in this function:
  - CPL_ERROR_NULL_INPUT if (one of) the input pointer(s) is NULL
 */
/*----------------------------------------------------------------------------*/
int cpl_imagelist_is_uniform(const cpl_imagelist * imlist)
{
    cpl_type     type;
    int          nx, ny;
    int          i;
    
    cpl_ensure(imlist != NULL, CPL_ERROR_NULL_INPUT, -1);
    if (imlist->ni == 0) return 1;

    /* Check the images */
    nx   = cpl_image_get_size_x(imlist->images[0]);
    ny   = cpl_image_get_size_y(imlist->images[0]);
    type = cpl_image_get_type  (imlist->images[0]);

    for (i=1 ; i < imlist->ni ; i++) {
        if (cpl_image_get_size_x(imlist->images[i]) != nx)   return i+1;
        if (cpl_image_get_size_y(imlist->images[i]) != ny)   return i+1;
        if (cpl_image_get_type  (imlist->images[i]) != type) return i+1;
    }
    return 0 ;
}

/**@}*/
