/* $Id: cpl_tools.c,v 1.64 2007/01/17 13:30:06 llundin Exp $
 *
 * This file is part of the ESO Common Pipeline Library
 * Copyright (C) 2001-2004 European Southern Observatory
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/*
 * $Author: llundin $
 * $Date: 2007/01/17 13:30:06 $
 * $Revision: 1.64 $
 * $Name: cpl-3_1 $
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

/*-----------------------------------------------------------------------------
                                   Includes
 -----------------------------------------------------------------------------*/

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <float.h>
#include <math.h>
#include <assert.h>
#include "time.h"

#include <cxmemory.h>
#include <cxmessages.h>
#include <cxutils.h>

#include "cpl_propertylist_impl.h"
#include "cpl_memory.h"
#include "cpl_tools.h"

/*-----------------------------------------------------------------------------
                                   Defines
 -----------------------------------------------------------------------------*/

/* Swap macros */
#define CPL_DOUBLE_SWAP(a,b) { register double t=(a);(a)=(b);(b)=t; }
#define CPL_FLOAT_SWAP(a,b) { register float t=(a);(a)=(b);(b)=t; }
#define CPL_INT_SWAP(a,b) { register int t=(a);(a)=(b);(b)=t; }

static cpl_flops flop_count = 0;

/*-----------------------------------------------------------------------------
                        Private function prototypes
 -----------------------------------------------------------------------------*/

static double cpl_tools_get_median_9double(double *) ;

/*----------------------------------------------------------------------------*/
/**
 * @defgroup cpl_tools Utility functions
 *
 * This module provides various functions to apply simple operations on
 * data arrays (sorting, median computation) and a function to monitor the
 * CPU time.
 *
 * @par Synopsis:
 * @code
 *   #include "cpl_tools.h"
 * @endcode
 */
/*----------------------------------------------------------------------------*/
/**@{*/

/*-----------------------------------------------------------------------------
                            Function codes
 -----------------------------------------------------------------------------*/

/*----------------------------------------------------------------------------*/
/**
  @brief    Get the kth smallest value in a double array
  @param    a   the double array
  @param    n   the array size
  @param    k   the requested value position in the sorted array
  @return   the kth smallest value in the sorted array.

  The input array is modified.
  In case of error, the #_cpl_error_code_ code is set, and the returned value
  is undefined.

  Possible #_cpl_error_code_ set in this function:
  - CPL_ERROR_NULL_INPUT
 */
/*----------------------------------------------------------------------------*/
double cpl_tools_get_kth_double(
        double  *   a,
        int         n,
        int         k)
{
    register double x ;
    register int    i, j, l, m ;

    cpl_ensure(a, CPL_ERROR_NULL_INPUT, 0.00) ;

    l=0 ; m=n-1 ;
    while (l<m) {
        x=a[k] ;
        i=l ;
        j=m ;
        do {
            while (a[i]<x) i++ ;
            while (x<a[j]) j-- ;
            if (i<=j) {
                CPL_DOUBLE_SWAP(a[i],a[j]) ;
                i++ ; j-- ;
            }
        } while (i<=j) ;
        if (j<k) l=i ;
        if (k<i) m=j ;
    }
    return a[k] ;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Get the kth smallest value in a float array
  @param    a   the float array
  @param    n   the array size
  @param    k   the requested value position in the sorted array
  @return   the kth smallest value in the sorted array.
  @see      cpl_tools_get_kth_double()
 */
/*----------------------------------------------------------------------------*/
float cpl_tools_get_kth_float(
        float   *   a,
        int         n,
        int         k)
{
    register float  x ;
    register int    i, j, l, m ;

    cpl_ensure(a, CPL_ERROR_NULL_INPUT, 0.00) ;

    l=0 ; m=n-1 ;
    while (l<m) {
        x=a[k] ;
        i=l ;
        j=m ;
        do {
            while (a[i]<x) i++ ;
            while (x<a[j]) j-- ;
            if (i<=j) {
                CPL_FLOAT_SWAP(a[i],a[j]) ;
                i++ ; j-- ;
            }
        } while (i<=j) ;
        if (j<k) l=i ;
        if (k<i) m=j ;
    }
    return a[k] ;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Get the kth smallest value in an integer array
  @param    a   the integer array
  @param    n   the array size
  @param    k   the requested value position in the sorted array
  @return   the kth smallest value in the sorted array.
  @see      cpl_tools_get_kth_double()
 */
/*----------------------------------------------------------------------------*/
int cpl_tools_get_kth_int(
        int     *   a,
        int         n,
        int         k)
{
    register float  x ;
    register int    i, j, l, m ;

    cpl_ensure(a, CPL_ERROR_NULL_INPUT, 0) ;

    l=0 ; m=n-1 ;
    while (l<m) {
        x=a[k] ;
        i=l ;
        j=m ;
        do {
            while (a[i]<x) i++ ;
            while (x<a[j]) j-- ;
            if (i<=j) {
                CPL_INT_SWAP(a[i],a[j]) ;
                i++ ; j-- ;
            }
        } while (i<=j) ;
        if (j<k) l=i ;
        if (k<i) m=j ;
    }
    return a[k] ;
}

#define CPL_PIX_STACK_SIZE 50
/*----------------------------------------------------------------------------*/
/**
  @brief    Sort a double array
  @param    pix_arr     the array to sort
  @param    n           the array size
  @return   the #_cpl_error_code_ or CPL_ERROR_NONE

  Possible #_cpl_error_code_ set in this function:
  - CPL_ERROR_NULL_INPUT
  - CPL_ERROR_ILLEGAL_INPUT
 */
/*----------------------------------------------------------------------------*/
cpl_error_code cpl_tools_sort_double(
        double  *   pix_arr,
        int         n)
{
    int         i, ir, j, k, l;
    int     *   i_stack ;
    int         j_stack ;
    double      a ;

    /* Check entries */
    cpl_ensure(pix_arr, CPL_ERROR_NULL_INPUT,
            CPL_ERROR_NULL_INPUT) ;

    ir = n ;
    l = 1 ;
    j_stack = 0 ;
    i_stack = cpl_malloc(CPL_PIX_STACK_SIZE * sizeof(double)) ;
    for (;;) {
        if (ir-l < 7) {
            for (j=l+1 ; j<=ir ; j++) {
                a = pix_arr[j-1];
                for (i=j-1 ; i>=1 ; i--) {
                    if (pix_arr[i-1] <= a) break;
                    pix_arr[i] = pix_arr[i-1];
                }
                pix_arr[i] = a;
            }
            if (j_stack == 0) break;
            ir = i_stack[j_stack-- -1];
            l  = i_stack[j_stack-- -1];
        } else {
            k = (l+ir) >> 1;
            CPL_DOUBLE_SWAP(pix_arr[k-1], pix_arr[l])
            if (pix_arr[l] > pix_arr[ir-1]) {
                CPL_DOUBLE_SWAP(pix_arr[l], pix_arr[ir-1])
            }
            if (pix_arr[l-1] > pix_arr[ir-1]) {
                CPL_DOUBLE_SWAP(pix_arr[l-1], pix_arr[ir-1])
            }
            if (pix_arr[l] > pix_arr[l-1]) {
                CPL_DOUBLE_SWAP(pix_arr[l], pix_arr[l-1])
            }
            i = l+1;
            j = ir;
            a = pix_arr[l-1];
            for (;;) {
                do i++; while (pix_arr[i-1] < a);
                do j--; while (pix_arr[j-1] > a);
                if (j < i) break;
                CPL_DOUBLE_SWAP(pix_arr[i-1], pix_arr[j-1]);
            }
            pix_arr[l-1] = pix_arr[j-1];
            pix_arr[j-1] = a;
            j_stack += 2;
            if (j_stack > CPL_PIX_STACK_SIZE) {
                /* Should never reach here */
                cpl_free(i_stack);
                return CPL_ERROR_ILLEGAL_INPUT ;
            }
            if (ir-i+1 >= j-l) {
                i_stack[j_stack-1] = ir;
                i_stack[j_stack-2] = i;
                ir = j-1;
            } else {
                i_stack[j_stack-1] = j-1;
                i_stack[j_stack-2] = l;
                l = i;
            }
        }
    }
    cpl_free(i_stack) ;
    return CPL_ERROR_NONE ;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Sort an integer array
  @param    pix_arr     the array to sort
  @param    n           the array size
  @return   the #_cpl_error_code_ or CPL_ERROR_NONE

  Possible #_cpl_error_code_ set in this function:
  - CPL_ERROR_NULL_INPUT
  - CPL_ERROR_ILLEGAL_INPUT
 */
/*----------------------------------------------------------------------------*/
cpl_error_code cpl_tools_sort_int(
        int     *   pix_arr,
        int         n)
{
    int         i, ir, j, k, l;
    int *       i_stack ;
    int         j_stack ;
    int         a ;

    /* Check entries */
    cpl_ensure(pix_arr, CPL_ERROR_NULL_INPUT,
            CPL_ERROR_NULL_INPUT) ;

    ir = n ;
    l = 1 ;
    j_stack = 0 ;
    i_stack = cpl_malloc(CPL_PIX_STACK_SIZE * sizeof(double)) ;
    for (;;) {
        if (ir-l < 7) {
            for (j=l+1 ; j<=ir ; j++) {
                a = pix_arr[j-1];
                for (i=j-1 ; i>=1 ; i--) {
                    if (pix_arr[i-1] <= a) break;
                    pix_arr[i] = pix_arr[i-1];
                }
                pix_arr[i] = a;
            }
            if (j_stack == 0) break;
            ir = i_stack[j_stack-- -1];
            l  = i_stack[j_stack-- -1];
        } else {
            k = (l+ir) >> 1;
            CPL_INT_SWAP(pix_arr[k-1], pix_arr[l])
            if (pix_arr[l] > pix_arr[ir-1]) {
                CPL_INT_SWAP(pix_arr[l], pix_arr[ir-1])
            }
            if (pix_arr[l-1] > pix_arr[ir-1]) {
                CPL_INT_SWAP(pix_arr[l-1], pix_arr[ir-1])
            }
            if (pix_arr[l] > pix_arr[l-1]) {
                CPL_INT_SWAP(pix_arr[l], pix_arr[l-1])
            }
            i = l+1;
            j = ir;
            a = pix_arr[l-1];
            for (;;) {
                do i++; while (pix_arr[i-1] < a);
                do j--; while (pix_arr[j-1] > a);
                if (j < i) break;
                CPL_INT_SWAP(pix_arr[i-1], pix_arr[j-1]);
            }
            pix_arr[l-1] = pix_arr[j-1];
            pix_arr[j-1] = a;
            j_stack += 2;
            if (j_stack > CPL_PIX_STACK_SIZE) {
                /* Should never reach here */
                cpl_free(i_stack);
                return CPL_ERROR_ILLEGAL_INPUT ;
            }
            if (ir-i+1 >= j-l) {
                i_stack[j_stack-1] = ir;
                i_stack[j_stack-2] = i;
                ir = j-1;
            } else {
                i_stack[j_stack-1] = j-1;
                i_stack[j_stack-2] = l;
                l = i;
            }
        }
    }
    cpl_free(i_stack) ;
    return CPL_ERROR_NONE ;
}
#undef CPL_PIX_STACK_SIZE

/*----------------------------------------------------------------------------*/
/**
  @brief    Compute the sample variance of a double array
  @param    a     The double array
  @param    n     The non-negative array size
  @param    mean  The sample mean of a
  @return   The variance

  The sample variance is
    S(n) = (1/n) sum((x_i-mean)^2) (i=1 -> n)

  mean may be different from the actual mean, in which case something else is
  computed.
 */
/*----------------------------------------------------------------------------*/
double cpl_tools_get_variance_double(const double * a, int n, double mean)
{
    double variance = 0;
    int i;

    cpl_ensure(a != NULL, CPL_ERROR_NULL_INPUT, 0.0);
    cpl_ensure(n >= 0, CPL_ERROR_ILLEGAL_INPUT, 0.0);

  /* Compute the variance of the elements using the recurrence relation
     var_n = var_{n-1} + ((x_n-mean)*(x_n-mean) - var_{n-1})/(n+1) */

    for (i=0; i < n; i++) {
        const double delta = a[i] - mean;
        variance += (delta * delta - variance) / (i + 1);
    }
    return variance;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Get the median value in a double array
  @param    a   the double array
  @param    n   the array size
  @return   the requested median
  @see      cpl_tools_get_kth_double()

  For a finite population or sample, the median is the middle value of an odd
  number of values (arranged in ascending order) or any value between the two
  middle values of an even number of values.
  The computed median should actually be a value of the input array, so, in
  case of an even number of values in the input array, the median would be the
  lowest of the two central values.

  The function is optimized for 9 elements.

  Example: the median of (1 2 3) is 2 and the median of (1 2 3 4) is 2.
 */
/*----------------------------------------------------------------------------*/
double cpl_tools_get_median_double(
        double  *   a,
        int         n)
{
    if (n==9) return cpl_tools_get_median_9double(a) ;
    return cpl_tools_get_kth_double(a,n,(((n)&1)?((n)/2):(((n)/2)-1))) ;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Get the median value in a float array
  @param    a   the float array
  @param    n   the array size
  @return   the requested median
  @see      cpl_tools_get_median_double()
 */
/*----------------------------------------------------------------------------*/
float cpl_tools_get_median_float(
        float   *   a,
        int         n)
{
    return cpl_tools_get_kth_float(a,n,(((n)&1)?((n)/2):(((n)/2)-1))) ;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Get the median value in an integer array
  @param    a   the integer array
  @param    n   the array size
  @return   the requested median
  @see      cpl_tools_get_median_double()
 */
/*----------------------------------------------------------------------------*/
int cpl_tools_get_median_int(
        int     *   a,
        int         n)
{
    return cpl_tools_get_kth_int(a,n,(((n)&1)?((n)/2):(((n)/2)-1))) ;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Timer handling, including benchmarking.
  @param    mode [CPL_CLOCK_START CPL_CLOCK_STOP CPL_CLOCK_TIME CPL_CLOCK_RESET]
  @return   An elapsed time in seconds as a double.

  This routine is useful to benchmark the time spent in an image processing
  routine. It is used in three steps:

  1) Start the timer by calling @c cpl_tools_get_cputime(CPL_CLOCK_START)
  2) Stop the timer by calling @c cpl_tools_get_cputime(CPL_CLOCK_STOP)
  3) Get the total time by calling @c cpl_tools_get_cputime(CPL_CLOCK_TIME)

  Steps 1) and 2) may be repeated pairwise more than once.
  After any instance of step 2) step 3) may be repeated more than once.
  cpl_tools_get_cputime(CPL_CLOCK_TIME) may also be called before step 1),
  in which case it returns 0.

  The call with CPL_CLOCK_START returns the total elapsed time in seconds since
  the begining.
  The call with CPL_CLOCK_STOP returns the time elapsed since the last
  CPL_CLOCK_START call.
  The call with CPL_CLOCK_TIME return the total time elapsed between all
  previous CPL_CLOCK_START - CPL_CLOCK_STOP calls.

  Additionally, the call with CPL_CLOCK_RESET sets the internal state of the
  timing function to its initial state, something that may be necessary when
  recovering from an error. In this case the function returns zero.

  Possible #_cpl_error_code_ set in this function:
  - CPL_ERROR_ILLEGAL_INPUT
 */
/*----------------------------------------------------------------------------*/
double cpl_tools_get_cputime(cpl_clock mode)
{
    /* The state of the static variables */
    static cpl_clock pmode  = CPL_CLOCK_STOP;

    /* clock() may wrap around after as little as 36 minutes,
       so store the result in a double */
    static clock_t chrono = 0;
    static double  total  = 0;
    clock_t        diff;
    double         result ;

    switch (mode) {
        case CPL_CLOCK_START:
            cpl_ensure(pmode != CPL_CLOCK_START, CPL_ERROR_ILLEGAL_INPUT, 0);
            chrono = clock() ;
            result = chrono;
            pmode = mode;
            break ;
        case CPL_CLOCK_STOP:
            cpl_ensure(pmode == CPL_CLOCK_START, CPL_ERROR_ILLEGAL_INPUT, 0);
            /* At this point clock() may have wrapped around, causing it to be
               much smaller than chrono. Subtraction must therefore be done in
               type (clock_t). */
            diff = clock() - chrono;
            result = diff;
            total += result;
            pmode = mode;
            break ;
        case CPL_CLOCK_TIME:
            cpl_ensure(pmode != CPL_CLOCK_START, CPL_ERROR_ILLEGAL_INPUT, 0);
            result = total ;
            break ;
        case CPL_CLOCK_RESET:
            /* Set the static variables to their initial values */
            pmode = CPL_CLOCK_STOP;
            chrono = 0;
            total = 0.0;
            result = 0.0; /* Need to return something */
            break ;
        default:
            /* Only reached if cpl_clock is extended in error */
            cpl_ensure(0, CPL_ERROR_UNSUPPORTED_MODE, 0.0);
    }

    return result / CLOCKS_PER_SEC;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Determine if an integer is a power of 2.
  @param    p   Integer to check.
  @return   The corresponding power of 2 or -1 if p is not a power of 2
 */
/*----------------------------------------------------------------------------*/
int cpl_tools_is_power_of_2(int p)
{
    int ipow = 0;
    int done;

    if (p <= 0) return -1;

    /* Count until 1st non-zero bit is found */
    do {
        done = p & 1;
        p >>= 1;
        ipow++;
    } while (!done);

    /* The number is a power iff p is now zero */
    return p == 0 ? ipow-1 : -1;

}

/*----------------------------------------------------------------------------*/
/**
  @brief   Compute x to the power of y
  @param   x  The base
  @param   y  The power
  @return  x to the power of y
  @note    Iff y is negative and x is zero an actual division by zero will occur

  Apart from a possible difference in round-off the result equals pow(x,y).

 */
/*----------------------------------------------------------------------------*/
double cpl_tools_ipow(double x, int y)
{

    double result;
    double pow2 = x;
    int p = abs(y);

    /* Compute the result as a product of powers of 2 of x.
       - this method may produce (slightly) different results than pow(x,y) */

    /* Handle least significant bit in p here in order to avoid an unnecessary
       multiplication of pow2 - which could cause an over- or underflow */
    result = p & 1 ? x : 1.0;

    while (p >>= 1) {
        pow2 *= pow2;
        /* Process least significant bit in p */
        if (p & 1) result *= pow2;
    }

    /* It is left to the caller to ensure that no division by zero occurs */
    return y < 0 ? 1.0/result : result;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Generate a valid FITS header for saving functions
  @param    qh_in       The valid minimal qfits_header
  @param    plist_in    The propertylist with the keywords to add
  @param    mode        0 for a primary header, 1 for an extension header
  @return   1 newly allocated valid qfits_header to use for saving

  If plist_in is NULL, the returned header is a copy if the input one.

  Possible #_cpl_error_code_ set in this function:
  - CPL_ERROR_ILLEGAL_INPUT
  - CPL_ERROR_NULL_INPUT if qh_in is NULL
 */
/*----------------------------------------------------------------------------*/
qfits_header * cpl_tools_gen_valid_header(
        const qfits_header      *   qh_in,
        const cpl_propertylist  *   plist_in,
        int                         mode)
{
    qfits_header        *   qh_out ;

    /* Test entries */
    cpl_ensure(qh_in, CPL_ERROR_NULL_INPUT, NULL) ;
    cpl_ensure(mode==1 || mode==0, CPL_ERROR_ILLEGAL_INPUT, NULL) ;

    /* Nothing to merge */
    if (plist_in == NULL) {
        qh_out = qfits_header_copy((qfits_header*)qh_in) ;
    } else {

#define CPL_TOOLS_VALID_HEADER_RM_REGEXP "SIMPLE|BSCALE|BZERO|EXTEND" \
        "|XTENSION|BITPIX|NAXIS[0-9]*|PCOUNT|GCOUNT" \
        "|TFIELDS|TFORM[0-9]*|TTYPE[0-9]*|TUNIT[0-9]*|TZERO[0-9]*|TSCAL[0-9]*"

        const char       * rmregexp;
        cpl_propertylist * plist = cpl_propertylist_new();
        qfits_header     * qsheader ;
        char               key[FITS_LINESZ+1];
        char               val[FITS_LINESZ+1];
        char               com[FITS_LINESZ+1];
        char               lin[FITS_LINESZ+1];
        int                n, i;


        if (mode == 0) {
            /* Primary header */
            rmregexp =  "^(" CPL_TOOLS_VALID_HEADER_RM_REGEXP 
                "|BLOCKED|EXTNAME|EXTVER|EXTLEVEL" ")$";

        } else {
            /* Extension header */
            rmregexp =  "^(" CPL_TOOLS_VALID_HEADER_RM_REGEXP 
                "|ORIGIN[0-9]*|DATE" ")$";
        }

        /* Copy all but the black-listed properties */
        if (cpl_propertylist_copy_property_regexp(plist, plist_in, rmregexp,
                                                  1)) {
            /* Reaching this point means a bug in CPL */
            cpl_propertylist_delete(plist);
            cpl_ensure(0, cpl_error_get_code(), NULL);
        }

        qsheader = cpl_propertylist_to_fits(plist);
        cpl_propertylist_delete(plist);

        cpl_ensure(qsheader, CPL_ERROR_ILLEGAL_INPUT, NULL) ;

        /* Duplicate the input header */
        qh_out = qfits_header_copy((qfits_header *)qh_in) ;

        n = qsheader->n - 1;             /* To avoid final "END" */
        for (i = 0; i < n; i++) {
            qfits_header_getitem(qsheader, i, key, val, com, lin);
            qfits_header_add(qh_out, key, val, com, lin);
        }
        qfits_header_destroy(qsheader);
    }
    return qh_out ;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Set the FLOPS counter to zero.
  @return   void
  @note This function is intended to be used only by the CPL test modules.

 */
/*----------------------------------------------------------------------------*/
void cpl_tools_reset_flops (void)
{
    flop_count = 0;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Add to the FLOPS counter.
  @param flops  FLOPS to add
  @return   void
  @note This function is intended to be used only by the CPL test modules.

 */
/*----------------------------------------------------------------------------*/
void cpl_tools_add_flops(cpl_flops flops)
{
    flop_count += flops;

}

/*----------------------------------------------------------------------------*/
/**
  @brief    Get the FLOPS count.
  @return   The FLOPS count
  @note This function is intended to be used only by the CPL test modules.

 */
/*----------------------------------------------------------------------------*/
cpl_flops cpl_tools_get_flops(void)
{
  return flop_count;
}

/*----------------------------------------------------------------------------*/
/**
   @brief
     Solution of Symmetric, Positive Definite system of linear equations.
 *
   @param self   The N by N Symmetric, Positive Definite matrix.
   @param rhs    An N by K matrix containing K right-hand-sides.
   @note rhs may contain more than one column,
         which each represent an independent right-hand-side. The solution
         is written in rhs. Only the upper triangular part of self is read,
         while the lower triangular part is modified.
   @return CPL_ERROR_NONE on success, or the relevant CPL error code
 *
   @error
     <table class="ec" align="center">
       <tr>
         <td class="ecl">CPL_ERROR_NULL_INPUT</td>
         <td class="ecr">
           <i>self</i> or <i>rhs</i> is a <tt>NULL</tt> pointer.
         </td>
       </tr>
       <tr>
         <td class="ecl">CPL_ERROR_INCOMPATIBLE_INPUT</td>
         <td class="ecr">
           <i>self</i> and <i>rhs</i> do not have the same number of rows.
         </td>
       </tr>
       <tr>
         <td class="ecl">CPL_ERROR_ILLEGAL_INPUT</td>
         <td class="ecr">
           <i>self</i> is not an N by N matrix.
         </td>
       </tr>
       <tr>
         <td class="ecl">CPL_ERROR_SINGULAR_MATRIX</td>
         <td class="ecr">
           <i>self</i> is (near) singular and a solution cannot be computed.
         </td>
       </tr>
     </table>
   @enderror
 */
/*----------------------------------------------------------------------------*/

cpl_error_code cpl_matrix_solve_spd(cpl_matrix *self,
                                    cpl_matrix *rhs)
{

    cpl_ensure_code(!cpl_matrix_decomp_chol(self), cpl_error_get_code());

    cpl_ensure_code(!cpl_matrix_solve_chol(self, rhs), cpl_error_get_code());

    return CPL_ERROR_NONE;

}


/**
 * @brief Create and compute A = B * transpose(B)
 *
 * @param self     M x N Matrix
 * @return Pointer to created M x M product matrix, or @c NULL on error.
 * @note Only the upper triangle of A is computed, while the elements
 *       below the main diagonal have undefined values.
 * @see cpl_matrix_product_create()
 *
 * @error
 *   <table class="ec" align="center">
 *     <tr>
 *       <td class="ecl">CPL_ERROR_NULL_INPUT</td>
 *       <td class="ecr">
 *         Any input matrix is a <tt>NULL</tt> pointer.
 *       </td>
 *     </tr>
 *   </table>
 * @enderror
 *
 * To destroy the new matrix the function @c cpl_matrix_delete() should
 * be used.
 */

cpl_matrix * cpl_matrix_product_normal_create(const cpl_matrix * self)
{

    cpl_matrix   * product;
    const double * aread = cpl_matrix_get_data(self);
    double       * bwrite;
    double         sum;
    const int      m = cpl_matrix_get_nrow(self);
    const int      n = cpl_matrix_get_ncol(self);
    int            i, j, k;


    cpl_ensure(self != NULL, CPL_ERROR_NULL_INPUT, NULL);

    product = cpl_matrix_new(m, m);

    bwrite = cpl_matrix_get_data(product);

    /* The result at (i,j) is the dot-product of i'th and j'th row */
    for (i = 0; i < m; i++) {
        for (j = i; j < m; j++) {
            sum = 0.0;
            for (k = 0; k < n; k++) {
                sum += aread[n * i + k] * aread[n * j + k];
            }
            bwrite[m * i + j] = sum;
        }
    }

    cpl_tools_add_flops(n * m * (m + 1));

    return product;

}


/*----------------------------------------------------------------------------*/
/**
  @brief  Create a string and fill it using vsnprintf()
  @param  format  The format string
  @param  arglist The argument list for the format
  @param  argcopy The argument list for the format - again
  @return The created string or NULL on error
  @note   The created string must be deallocated with cpl_free()
  @note   This function may only be called via cpl_vsprintf()
  @see vsnprintf()

  Possible CPL error codes set in this function:
  - CPL_ERROR_NULL_INPUT if the format string is NULL
  - CPL_ERROR_ILLEGAL_INPUT if the format string has an invalid format

 */
/*----------------------------------------------------------------------------*/
char * cpl_vsprintf_macro(const char * format, va_list arglist,
                             va_list argcopy)
{

    char   * self;
    size_t   selfsize;
    int      nlen;


    cpl_ensure(format != NULL, CPL_ERROR_NULL_INPUT, NULL);

    /* A guess at the necessary length of the string */
    selfsize = 1 + strlen(format);
    self = (char*)cpl_malloc(selfsize);

    nlen = cx_vsnprintf(self, (cxsize)selfsize, format, arglist);

    /* The format string may be invalid */
    if (nlen < 0) {
        cpl_free(self);
        cpl_ensure(0, CPL_ERROR_ILLEGAL_INPUT, NULL);
    }

    if (selfsize < 1 + (size_t)nlen) {
        /* self was too short, extend it */

        cpl_free(self);

         /* The exact size of self */
        selfsize = 1 + (size_t)nlen;

        self = (char*)cpl_malloc(selfsize);

        nlen = cx_vsnprintf(self, (cxsize)selfsize, format, argcopy);

        assert( nlen > 0);
        assert( selfsize == 1 + (size_t)nlen);

    } else if (selfsize > 1 + (size_t)nlen) {
        /* self was too long, shorten it */

        /* The exact size of self */
        selfsize = 1 + (size_t)nlen;

        self = (char*)cpl_realloc((void*)self, selfsize);

    }

    assert(selfsize == 1 + strlen(self));

    return self;

}

/*----------------------------------------------------------------------------*/
/**
  @brief  Create a new string by concatenation
  @param  format    The format string
  @param  ...       Variable argument list for format
  @return The allocated string with the concatenation or NULL on error
  @note   The returned string must be deallocated with cpl_free()
  @see cpl_vsprintf()

  Example of usage:
   @code

       int error;

       char * cp_cmd = cpl_sprintf("cp %s %s/%s", long_file, new_dir,
                                                     new_file);
       assert( cp_cmd != NULL);

       error = system(cp_cmd);

       assert(!error);

       cpl_free(cp_cmd);
                       
   @endcode


  Possible CPL error codes set in this function:
  - CPL_ERROR_NULL_INPUT if the format string is NULL
  - CPL_ERROR_ILLEGAL_INPUT if the format string has an invalid format

 */
/*----------------------------------------------------------------------------*/
char * cpl_sprintf(const char * format, ...)
{

    char    * self;
    va_list   arglist;
    va_list   argcopy;


    cpl_ensure(format != NULL, CPL_ERROR_NULL_INPUT, NULL);

    va_start(arglist, format);
    va_start(argcopy, format);
    self = cpl_vsprintf_macro(format, arglist, argcopy);
    va_end(arglist);
    va_end(argcopy);

    cpl_ensure(self != NULL, cpl_error_get_code(), NULL);

    return self;

}

/**@}*/

#define CPL_PIX_SORT(a,b) { if ((a)>(b)) CPL_DOUBLE_SWAP((a),(b)); }
/*----------------------------------------------------------------------------*/
/**
  @brief    Optimized median computation for 9 values arrays
  @param    p   array to sort
  @return   the requested median
  @see      cpl_tools_get_median_double()

  The input array is modified.
  In case of error, the #_cpl_error_code_ code is set, and the returned value
  is undefined.

  Possible #_cpl_error_code_ set in this function:
  - CPL_ERROR_NULL_INPUT
 */
/*----------------------------------------------------------------------------*/
static double cpl_tools_get_median_9double(double * p)
{
    cpl_ensure(p, CPL_ERROR_NULL_INPUT, 0.00) ;

    CPL_PIX_SORT(p[1], p[2]); CPL_PIX_SORT(p[4], p[5]); CPL_PIX_SORT(p[7],p[8]);
    CPL_PIX_SORT(p[0], p[1]); CPL_PIX_SORT(p[3], p[4]); CPL_PIX_SORT(p[6],p[7]);
    CPL_PIX_SORT(p[1], p[2]); CPL_PIX_SORT(p[4], p[5]); CPL_PIX_SORT(p[7],p[8]);
    CPL_PIX_SORT(p[0], p[3]); CPL_PIX_SORT(p[5], p[8]); CPL_PIX_SORT(p[4],p[7]);
    CPL_PIX_SORT(p[3], p[6]); CPL_PIX_SORT(p[1], p[4]); CPL_PIX_SORT(p[2],p[5]);
    CPL_PIX_SORT(p[4], p[7]); CPL_PIX_SORT(p[4], p[2]); CPL_PIX_SORT(p[6],p[4]);
    CPL_PIX_SORT(p[4], p[2]); return(p[4]) ;
}
#undef CPL_PIX_SORT
#undef CPL_DOUBLE_SWAP
#undef CPL_FLOAT_SWAP
#undef CPL_INT_SWAP
