/* $Id: cpl_type.c,v 1.2 2005/02/01 15:28:50 rpalsa Exp $
 *
 * This file is part of the ESO Common Pipeline Library
 * Copyright (C) 2001-2005 European Southern Observatory
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/*
 * $Author: rpalsa $
 * $Date: 2005/02/01 15:28:50 $
 * $Revision: 1.2 $
 * $Name: cpl-3_1 $
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <cxtypes.h>

#include "cpl_type.h"


/**
 * @defgroup cpl_types Type codes
 *
 * This module provides codes for the basic types (@c char, @c int, @c float,
 * etc.). These type codes may be used to indicate the type of a value stored
 * in another object, the value of a property or the pixel of an image
 * for instance. In addition, a utility function is provided to compute
 * the size, which is required to store a value of the type indicated by a
 * given type code.
 * 
 * The module 
 * @par Synopsis
 * @code
 *   #include <cpl_type.h>
 * @endcode
 */

/**@{*/

/**
 * @brief
 *   Compute the size of a type.
 *
 * @param type  Type code to be evaluated.
 *
 * @return The size of the fundamental type, or 0 in case an invalid type
 *   code was given.
 *
 * The function computes the atomic size of the type @em type. The result
 * for fundamental types like @c CPL_TYPE_FLOAT is what you would expect
 * from the C @b sizeof() operator. For arrays, i.e. types having the 
 * @c CPL_TYPE_FLAG_ARRAY set the returned size is not the size of a pointer
 * to @c CPL_TYPE_FLOAT for instance, but the size of its fundamental type,
 * i.e. the returned size is same as for the type @c CPL_TYPE_FLOAT.
 *
 * Especially for the type @c CPL_TYPE_STRING, which is explicitly defined for
 * convenience reasons, the size returned by this function is the size of
 * @c CPL_TYPE_CHAR!
 */

size_t 
cpl_type_get_sizeof(cpl_type type)
{

    cxsize sz;


    if (type & CPL_TYPE_INVALID)
        return 0;

    /*
     * We just want the size of the fundamental type, therefore the
     * array flag is removed. This also means that an explicit case
     * for CPL_TYPE_STRING is not necessary in the following switch.
     */

    type &= ~CPL_TYPE_FLAG_ARRAY;

    switch (type) {
        case CPL_TYPE_CHAR:
            sz = sizeof(cxchar);
            break;

        case CPL_TYPE_UCHAR:
            sz = sizeof(cxuchar);
            break;

        case CPL_TYPE_BOOL:
            sz = sizeof(cxbool);
            break;

        case CPL_TYPE_INT:
            sz = sizeof(cxint);
            break;

        case CPL_TYPE_UINT:
            sz = sizeof(cxuint);
            break;

        case CPL_TYPE_LONG:
            sz = sizeof(cxlong);
            break;

        case CPL_TYPE_ULONG:
            sz = sizeof(cxulong);
            break;

        case CPL_TYPE_FLOAT:
            sz = sizeof(cxfloat);
            break;

        case CPL_TYPE_DOUBLE:
            sz = sizeof(cxdouble);
            break;

        case CPL_TYPE_POINTER:
            sz = sizeof(cxptr);
            break;

        default:
            sz = 0;
            break;
    }

    return sz;

}
/**@}*/
